# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Amazon S3 staging system."""

import binascii
import collections
import functools
import hashlib
import logging
import os
import os.path
import pwd
import re
import stat
import sys
import threading

import argh
import boto.s3.connection
import boto.s3.bucket
import boto.s3.key
import boto.s3.lifecycle

from libgiza.git import GitRepo
from giza.config.helper import fetch_config

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

LOGGER = logging.getLogger('giza.operations.stage')
REDIRECT_PAT = re.compile('^Redirect 30[1|2|3] (\S+)\s+(\S+)', re.M)
FileUpdate = collections.namedtuple('FileUpdate', ['path', 'file_hash'])
AuthenticationInfo = collections.namedtuple('AuthenticationInfo', ['access_key', 'secret_key', 'username'])

CONFIG_PATH = '~/.config/giza-aws-authentication.conf'
SAMPLE_CONFIG = '''[authentication]
accesskey=<AWS access key>
secretkey=<AWS secret key>
'''


class Path(object):
    """Wraps Unix-style paths to ensure a normalized format."""
    def __init__(self, init):
        self.segments = init.split('/')

    def replace_prefix(self, orig, new):
        """Replace the "orig" string in this path ONLY if it is at the start."""
        cur = str(self)
        if cur.startswith(orig):
            return Path(str(self).replace(orig, new, 1))

        return Path(str(self))

    def ensure_prefix(self, prefix):
        """Prepend a string to this path ONLY if it does not already exist."""
        cur = str(self)
        if cur.startswith(prefix):
            return Path(str(self))

        return Path('/'.join((prefix, cur)))

    def __str__(self):
        """Format this path as a Unix-style path string."""
        return '/'.join(self.segments)


class BulletProofS3(object):
    """An S3 API that wraps boto to work around Amazon's 100-request limit."""
    THRESHOLD = 20

    def __init__(self, access_key, secret_key, bucket_name):
        self.access_key = access_key
        self.secret_key = secret_key
        self.name = bucket_name

        # For "dry" runs; only pretend to do anything.
        self.dry_run = False

        self.times_used = 0
        self._conn = None

    def get_connection(self):
        """Return a connection to S3. Not idempotent: will create a new connection
           after some number of calls."""
        if not self._conn or self.times_used > self.THRESHOLD:
            conn = boto.s3.connection.S3Connection(self.access_key, self.secret_key)
            self._conn = conn.get_bucket(self.name)
            self.times_used = 0

        self.times_used += 1
        return self._conn

    def list(self, prefix=None):
        """Returns a list of keys in this S3 bucket."""
        return self.get_connection().list(prefix=prefix)

    def get_redirect(self, name):
        """Return the destination redirect associated with a given source path."""
        k = boto.s3.key.Key(self.get_connection(), name)
        return k.get_redirect()

    def set_redirect(self, key, dest):
        """Redirect a given key to a destination."""
        if not self.dry_run:
            key.set_redirect(dest)

    def delete_keys(self, keys):
        """Delete the given list of keys."""
        if self.dry_run:
            keys = []

        return self.get_connection().delete_keys(keys)

    def copy(self, src_path, dest_name, **options):
        """Copy a given key to the given path."""
        if self.dry_run:
            return

        key = boto.s3.key.Key(self.get_connection())
        key.key = src_path
        key.copy(self.name, dest_name, **options)

    def upload_path(self, src, dest, **options):
        key = boto.s3.key.Key(self.get_connection())
        key.key = dest

        if self.dry_run:
            return key

        key.set_contents_from_filename(src, **options)
        return key


class StagingException(Exception):
    """Base class for all giza stage exceptions."""
    pass


class NoSuchEdition(StagingException):
    """An exception indicating that the requested edition does not exist."""
    pass


class SyncFileException(StagingException):
    """An exception indicating an S3 deletion error."""
    def __init__(self, path, reason):
        StagingException.__init__(self, 'Error syncing path: {0}'.format(path))
        self.reason = reason
        self.path = path


class SyncException(StagingException):
    """An exception indicating an error uploading files."""
    def __init__(self, errors):
        StagingException.__init__(self, 'Errors syncing data')
        self.errors = errors


class PoolWorker(threading.Thread):
    """A threaded worker that works on a queue of functions."""
    def __init__(self, **kwargs):
        threading.Thread.__init__(self, **kwargs)

        # Avoid needing to mutex-protect this queue by only ever appending to it
        self.tasks = []
        self.results = []
        self.i = 0

        self.semaphore = threading.Semaphore(0)

    def run(self):
        while True:
            self.semaphore.acquire()

            next_task = self.tasks[self.i]
            if next_task is None:
                return

            try:
                result = next_task()
                if result is not None:
                    self.results.append((next_task, result))
            except Exception as err:
                self.results.append((next_task, err))

            self.i += 1

    def add_task(self, task):
        """Adds a function to this work queue. The worker will exit if it
           reaches a None value."""
        self.tasks.append(task)
        self.semaphore.release()


def run_pool(tasks, n_workers=10, retries=1):
    """Run a list of tasks using a pool of threads. Return non-None results or
       exceptions as a list of (task, result) pairs in an arbitrary order."""
    assert retries >= 0

    results = []
    while retries >= 0:
        workers = []
        for _ in range(n_workers):
            worker = PoolWorker()
            worker.start()
            workers.append(worker)

        # Schedule tasks round-robin
        for i in range(len(tasks)):
            workers[i % n_workers].add_task(tasks[i])

        for worker in workers:
            worker.add_task(None)

        for worker in workers:
            worker.join()

        results = []
        for worker in workers:
            results.extend(worker.results)

        if not results:
            return []

        tasks = [result[0] for result in results]
        retries -= 1

    raise SyncException([result[1] for result in results])


def translate_htaccess(path):
    """Read a .htaccess file, and transform redirects into a mapping of redirects."""
    with open(path, 'r') as f:
        data = f.read()
        return dict([(x.lstrip('/'), y) for x, y in REDIRECT_PAT.findall(data)])

    return {}


class StagingCollector(object):
    """A dummy file collector interface that always reports files as having
       changed. Yields files and all symlinks.

       Obtain use_branch from StagingTargetConfig. If it is True,
       StagingCollector will only recurse into the directory given by branch."""
    def __init__(self, branch, use_branch, namespace):
        self.removed_files = []
        self.branch = branch
        self.use_branch = use_branch
        self.namespace = namespace

    def get_upload_set(self, root):
        return set(os.listdir(root))

    def collect(self, root, remote_keys):
        self.removed_files = []
        remote_hashes = {}
        whitelist = self.get_upload_set(root)

        LOGGER.info('Publishing %s', ', '.join(whitelist))

        # List all current redirects
        tasks = []
        for key in remote_keys:
            local_key = key.key.replace(self.namespace, '', 1)
            local_key = local_key.lstrip('/')

            # Don't register redirects for deletion in this stage
            if key.size == 0:
                continue

            # Check if we want to skip this path
            if local_key.split('/', 1)[0] not in whitelist:
                continue

            # Store its MD5 hash. Might be useless if encryption or multi-part
            # uploads are used.
            remote_hashes[local_key] = key.etag.strip('"')

            if not os.path.exists(os.path.join(root, local_key)):
                LOGGER.warn('Removing %s', os.path.join(root, local_key))
                self.removed_files.append(key.key)

        LOGGER.info('Running collection tasks')
        run_pool(tasks)
        LOGGER.info('Done. Scanning local filesystem')

        for basedir, dirs, files in os.walk(root, followlinks=True):
            # Skip branches we wish not to publish
            if basedir == root:
                dirs[:] = [d for d in dirs if d in whitelist]

            for filename in files:
                # Skip dotfiles
                if filename.startswith('.'):
                    continue

                path = os.path.join(basedir, filename)

                try:
                    local_hash = self.hash(path)
                except IOError:
                    continue

                remote_path = path.replace(root, '')
                if remote_hashes.get(remote_path, None) == local_hash:
                    continue

                yield FileUpdate(path, local_hash)

    def commit(self):
        pass

    def purge_now(self):
        pass

    @staticmethod
    def hash(path):
        """Return the cryptographic hash of the given file path as a byte
           sequence."""
        hasher = hashlib.md5()

        with open(path, 'rb') as input_file:
            while True:
                data = input_file.read(2**13)
                if not data:
                    break

                hasher.update(data)

        return hasher.hexdigest()


class DeployCollector(StagingCollector):
    def get_upload_set(self, root):
        if not self.use_branch:
            return set(os.listdir(root))

        # Special-case the root directory, because we want to publish only:
        # - Files
        # - The current branch (if published)
        # - Symlinks pointing to the current branch
        upload = set()
        for entry in os.listdir(root):
            path = os.path.join(root, entry)
            if os.path.isdir(path) and entry == self.branch:
                # This is the branch we want to upload
                upload.add(entry)
                continue

            # Only collect links that point to the current branch
            try:
                candidate = os.path.basename(os.path.realpath(path))
                if candidate == self.branch:
                    upload.add(entry)
                    continue
            except OSError:
                pass

        return upload


class Staging(object):
    # These files are always unique, and so there is no benefit in sharing them
    SHARED_CACHE_BLACKLIST = {'genindex.html', 'objects.inv', 'searchindex.js'}

    S3_OPTIONS = {'reduced_redundancy': True}
    RESERVED_BRANCHES = {'cache'}
    PAGE_SUFFIX = ''
    USE_CACHE = True
    EXPECT_HTACCESS = False

    FileCollector = StagingCollector

    def __init__(self, namespace, branch, staging_config, s3):
        # The S3 prefix for this staging site
        self.namespace = namespace
        self.branch = branch
        self.staging_config = staging_config

        self.s3 = s3
        self.collector = self.FileCollector(branch, staging_config.use_branch, namespace)

    @classmethod
    def compute_namespace(cls, prefix, username, branch, edition):
        """Staging places each stage under a unique namespace computed from an
           arbitrary username, branch, and edition. This helper returns such
           a namespace, appropriate for constructing a new Staging instance."""
        if branch in cls.RESERVED_BRANCHES:
            raise ValueError(branch)

        return '/'.join([x for x in (prefix, username, branch, edition) if x])

    def purge(self):
        """Remove all files associated with this branch and edition."""
        # Remove files from the index first; if the system dies in an
        # inconsistent state, we want to err on the side of reuploading too much
        self.collector.purge_now()

        prefix = '' if not self.namespace else '/'.join((self.namespace, ''))

        keys = [k.key for k in self.s3.list(prefix=prefix)]
        result = self.s3.delete_keys(keys)
        if result.errors:
            raise SyncException(result.errors)

    def stage(self, root):
        """Synchronize the build directory with the staging bucket under
           the namespace [username]/[branch]/[edition]/"""
        tasks = []

        redirects = {}
        htaccess_path = os.path.join(root, '.htaccess')
        try:
            if self.branch == 'master':
                redirects = translate_htaccess(htaccess_path)
        except IOError:
            # Only log an error if deploying; staging doesn't use .htaccess
            level = logging.CRITICAL if self.EXPECT_HTACCESS else logging.INFO
            LOGGER.log(level, 'Couldn\'t open required %s', htaccess_path)

            if self.EXPECT_HTACCESS:
                sys.exit(1)

        # Ensure that the root ends with a trailing slash to make future
        # manipulations more predictable.
        if not root.endswith('/'):
            root += '/'

        if not os.path.isdir(root):
            raise NoSuchEdition(root)

        # If a redirect is masking a file, we can run into an invalid 404
        # when the redirect is deleted but the file isn't republished.
        # If this is the case, warn and delete the redirect.
        for src, dest in redirects.items():
            src_path = os.path.join(root, src)
            if os.path.isfile(src_path) and os.path.basename(src_path) in os.listdir(os.path.dirname(src_path)):
                LOGGER.warn('Would ignore redirect that will mask file: %s', src)
#                del redirects[src]

        # Collect files that need to be uploaded
        for entry in self.collector.collect(root, self.s3.list(prefix=self.namespace)):
            src = entry.path.replace(root, '', 1)

            if os.path.islink(entry.path):
                # If redirecting from a directory, make sure we end it with a '/'
                suffix = self.PAGE_SUFFIX if os.path.isdir(entry.path) and not entry.path.endswith('/') else ''

                resolved = os.path.join(os.path.dirname(entry.path), os.readlink(entry.path))
                if os.path.islink(resolved):
                    LOGGER.warn('Multiple layers of symbolic link: %s', resolved)

                if not os.path.exists(resolved):
                    LOGGER.warn('Dead link: %s -> %s', entry.path, resolved)

                if not resolved.startswith(root):
                    LOGGER.warn('Skipping symbolic link %s: outside of root %s', resolved, root)

                redirects[str(Path(src + suffix).ensure_prefix(self.namespace))] = resolved.replace(root, '/', 1)
            else:
                tasks.append(functools.partial(
                    lambda path, file_hash: self.__upload(
                        path,
                        os.path.join(root, path),
                        file_hash),
                    src, entry.file_hash))

        # Run our actual staging operations in a thread pool. This would be
        # better with async IO, but this will do for now.
        LOGGER.info('Running %s tasks', len(tasks))
        run_pool(tasks)

        # XXX Right now we only sync redirects on master.
        #     Why: Master has the "canonical" .htaccess, and we'd need to attach
        #          metadata to each redirect on S3 to differentiate .htaccess
        #          redirects from symbolic links.
        #     Ramifications: Symbolic link redirects for non-master branches
        #                    will never be published.
        if self.branch == 'master':
            self.sync_redirects(redirects)

        # Remove from staging any files that our FileCollector thinks have been
        # deleted locally.
        remove_keys = [str(path.replace_prefix(root, '').ensure_prefix(self.namespace))
                       for path in [Path(p) for p in self.collector.removed_files]]

        if remove_keys:
            LOGGER.warn('Removing %s', remove_keys)
            remove_result = self.s3.delete_keys(remove_keys)
            if remove_result.errors:
                raise SyncException(remove_result.errors)

        self.collector.commit()

    def sync_redirects(self, redirects):
        """Upload the given path->url redirect mapping to the remote bucket."""
        if not redirects:
            return

        redirect_dirs = self.staging_config.redirect_dirs
        if not redirect_dirs:
            LOGGER.warn('No "redirect_dirs" listed for this project.')
            LOGGER.warn('No redirects will be removed.')

        LOGGER.info('Finding redirects to remove')
        removed = []
        remote_redirects = self.s3.list()
        for key in remote_redirects:
            # Make sure this is a redirect
            if key.size != 0:
                continue

            # Redirects are written /foo/bar/, not /foo/bar/index.html
            redirect_key = key.key.rsplit(self.PAGE_SUFFIX, 1)[0]

            # If it doesn't start with our namespace, ignore it
            if not redirect_key.startswith(self.namespace):
                continue

            # If it doesn't match one of our "owned" directories, ignore it
            if not [True for pat in redirect_dirs if pat.match(redirect_key)]:
                continue

            if redirect_key not in redirects:
                removed.append(key.key)

        LOGGER.warn('Removing %s redirects', len(removed))
        for remove in removed:
            LOGGER.warn('Removing redirect %s', remove)

        self.s3.delete_keys(removed)

        LOGGER.info('Creating %s redirects', len(redirects))
        tasks = []
        for src in redirects:
            tasks.append(functools.partial(
                    lambda src, dest, suffix: self.__redirect(
                        src + suffix,
                        dest),
                    src, redirects[src], self.PAGE_SUFFIX))
        run_pool(tasks)

    def __upload(self, local_path, src_path, file_hash):
        full_name = '/'.join((self.namespace, local_path))

        try:
            # If we don't have a hash, we can't do our normal caching stuff.
            # Just upload the file.
            if local_path in self.SHARED_CACHE_BLACKLIST or file_hash is None or not self.USE_CACHE:
                return self.__upload_nocache(src_path, full_name)

            # Try to copy the file from the cache to its destination
            file_hash = 'cache/' + binascii.b2a_hex(file_hash)

            try:
                self.s3.copy(file_hash, full_name, **self.S3_OPTIONS)
            except boto.exception.S3ResponseError as err:
                if err.status == 404:
                    # Not found in cache. Upload it to the cache
                    LOGGER.info('Uploading from %s to %s (%s)', local_path, full_name, file_hash)
                    self.s3.upload_path(src_path, file_hash, **self.S3_OPTIONS)

                    # And then copy it to its final destination
                    self.s3.copy(file_hash, full_name, **self.S3_OPTIONS)
                else:
                    raise err
        except boto.exception.S3ResponseError as err:
            raise SyncFileException(local_path, err.message)
        except IOError as err:
            LOGGER.exception('IOError while uploading file "%s": %s', local_path, err)

    def __upload_nocache(self, src_path, full_name):
        LOGGER.info('Uploading %s to %s', src_path, full_name)
        self.s3.upload_path(src_path, full_name, **self.S3_OPTIONS)

    def __redirect(self, src, dest):
        key = boto.s3.key.Key(self.s3.get_connection(), src)
        try:
            if key.get_redirect() == dest:
                LOGGER.debug('Skipping redirect %s', src)
                return
        except boto.exception.S3ResponseError as err:
            if err.status != 404:
                LOGGER.exception('S3 error creating redirect from %s to %s', src, dest)

        LOGGER.info('Redirecting %s to %s', src, dest)
        self.s3.set_redirect(key, dest)


class DeployStaging(Staging):
    PAGE_SUFFIX = '/index.html'
    USE_CACHE = False
    EXPECT_HTACCESS = True

    FileCollector = DeployCollector

    @classmethod
    def compute_namespace(cls, prefix, username, branch, edition):
        return '/'.join([x for x in (prefix, edition) if x])


def do_stage(root, staging):
    """Drive the main staging process for a single edition, and print nicer
       error messages for exceptions."""
    try:
        return staging.stage(root)
    except SyncException as err:
        LOGGER.error('Failed to upload some files:')
        for sub_err in err.errors:
            try:
                raise sub_err
            except SyncFileException:
                LOGGER.error('%s: %s', sub_err.path, sub_err.reason)
    except NoSuchEdition as err:
        LOGGER.error('No edition found at %s', err.message)
        LOGGER.info('Try specifying the -e [edition] option')


def create_config_framework(path):
    """Create a skeleton configuration file with appropriately locked-down
       permissions."""
    try:
        os.mkdir(os.path.dirname(path), 0o751)
    except OSError:
        pass

    # Make sure we don't write the framework if it already exists.
    try:
        with os.fdopen(os.open(path,
                               os.O_WRONLY | os.O_CREAT | os.O_EXCL,
                               0o600), 'wx') as conf_file:
            conf_file.write('[authentication]\n')
    except IOError:
        pass


class StagingPipeline(object):
    Staging = Staging

    def __init__(self, args, destage, dry_run=False):
        self.args = args
        self.conf = fetch_config(args)
        self.destage = destage
        self.dry_run = dry_run

        self.branch = GitRepo().current_branch()

    def check_builder(self):
        if not self.args.builder:
            print('No builder specified. Try specifying "-b html".')
            raise ValueError(self.args.builder)

    def setup_staging_conf(self):
        self.staging_config = self.conf.deploy.get_staging(self.conf.project.name)

    def load_authentication(self):
        """Returns an AuthenticationInfo instance, or None."""
        self.cfg_path = os.path.expanduser(CONFIG_PATH)
        cfg = configparser.ConfigParser()
        cfg.read(self.cfg_path)

        # Warn the user if config permissions are too lax
        try:
            if os.name == 'posix' and stat.S_IMODE(os.stat(self.cfg_path).st_mode) != 0o600:
                LOGGER.warn('Your AWS authentication file is poorly protected! You should run')
                LOGGER.warn('    chmod 600 %s', self.cfg_path)
        except OSError:
            pass

        # Load S3 authentication information
        try:
            access_key = cfg.get('authentication', 'accesskey')
            secret_key = cfg.get('authentication', 'secretkey')
        except (configparser.NoSectionError, configparser.NoOptionError):
            print('No staging authentication found. Create a file at {0} with '
                  'contents like the following:\n'.format(self.cfg_path))
            print(SAMPLE_CONFIG)
            create_config_framework(self.cfg_path)
            return None

        # Get the user's preferred name; we use this as part of our S3 namespaces
        try:
            username = cfg.get('personal', 'username')
        except (configparser.NoSectionError, configparser.NoOptionError):
            username = pwd.getpwuid(os.getuid())[0]

        self.auth = AuthenticationInfo(access_key, secret_key, username)
        self.s3 = BulletProofS3(self.auth.access_key,
                                self.auth.secret_key,
                                self.staging_config.bucket)
        self.s3.dry_run = self.dry_run

    def setup_sourcedir(self):
        self.editions = self.args.edition or ('',)
        edition_suffixes = [self.args.builder[0] + ('-' if edition else '') + edition
                            for edition in self.editions]

        self.roots = [os.path.join(self.conf.paths.projectroot,
                                   self.conf.paths.branch_output,
                                   edition_suffix) for edition_suffix in edition_suffixes]

    def stage(self):
        prefix = self.staging_config.prefix

        try:
            for root, edition in zip(self.roots, self.editions):
                namespace = self.Staging.compute_namespace(prefix, self.auth.username, self.branch, edition)
                staging = self.Staging(namespace, self.branch, self.staging_config, self.s3)

                if self.destage:
                    staging.purge()
                    continue

                do_stage(root, staging)
        except boto.exception.S3ResponseError as err:
            if err.status == 403:
                LOGGER.error('Failed to upload to S3: Permission denied.')
                LOGGER.info('Check your authentication configuration at %s, '
                            'and/or talk to IT.', self.cfg_path)
                return

            raise err

    def print_report(self):
        """Print a list of staging URLs corresponding to the given editions."""
        if not self.destage:
            print('Staged at:')
            for edition in self.editions:
                suffix = '/'.join((self.auth.username, self.branch))
                if edition:
                    suffix = '{0}/{1}/{2}'.format(self.auth.username, self.branch, edition)
                print('    {0}/{1}'.format(self.staging_config.url, suffix))

    def run(self):
        self.check_builder()
        self.setup_staging_conf()
        self.load_authentication()
        self.setup_sourcedir()
        self.stage()
        self.print_report()


class DeployPipeline(StagingPipeline):
    Staging = DeployStaging

    def setup_deploy_conf(self):
        self.staging_config = self.conf.deploy.get_deploy(self.conf.project.name)

    def setup_sourcedir(self):
        self.editions = self.args.edition or ('',)
        edition_suffixes = self.editions

        self.roots = [os.path.join(self.conf.paths.projectroot,
                                   self.conf.paths.public,
                                   edition_suffix) for edition_suffix in edition_suffixes]

    def print_report(self):
        """Print a list of staging URLs corresponding to the given editions."""
        print('Published to:')
        print('    ' + self.staging_config.url)

    def run(self):
        # Check if the active branch is publishable. This is a sanity check
        # more than anything else, since the active branch doesn't directly
        # correlate to what is built.
        if self.branch not in self.conf.git.branches.published:
            print('Current branch ({0}) is not published'.format(self.branch))
            return

        self.check_builder()
        self.setup_deploy_conf()
        self.load_authentication()
        self.setup_sourcedir()
        self.stage()
        self.print_report()


@argh.arg('--edition', '-e', nargs='*')
@argh.arg('--destage', default=False,
          dest='_destage', help='Delete the contents of the current staged render')
@argh.arg('--builder', '-b', default='html')
@argh.arg('--dry', default=False, dest='_dry')
@argh.named('stage')
@argh.expects_obj
def main_stage(args):
    """Start an HTTP server rooted in the build directory."""
    StagingPipeline(args, destage=args._destage, dry_run=args._dry).run()


@argh.arg('--edition', '-e', nargs='*')
@argh.arg('--builder', '-b', default='html')
@argh.arg('--dry', default=False, dest='_dry')
@argh.named('s3_deploy')
@argh.expects_obj
def main_deploy(args):
    """Start an HTTP server rooted in the build directory."""
    DeployPipeline(args, destage=False, dry_run=args._dry).run()



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Operations for managing and producing artifacts using the ``gettext`` builder,
``sphinx-intl``, and Transifex-based translation workflow.
"""

import logging
import os.path
import subprocess

import argh

import libgiza.app
import libgiza.task

from giza.config.helper import fetch_config

import sphinx_intl.commands

logger = logging.getLogger('giza.operations.tx')

# Helpers


class FileLogger(object):

    def __init__(self, logger, level=logging.INFO):
        self.logger = logger
        self.level = level

    def write(self, message):
        if message != '\n':
            self.logger.log(self.level, message)


def tx_resources(conf):
    tx_conf = os.path.join(conf.paths.projectroot,
                           ".tx", 'config')

    with open(tx_conf, 'r') as f:
        resources = [l.strip()[1:-1]
                     for l in f.readlines()
                     if l.startswith('[')][1:]

    return resources


def logged_command(verb, args):
    output = subprocess.check_output(args)
    logger.info('{0}ed {1}'.format(verb, args[-1]))

    return output


def check_for_orphaned_tx_files(conf):
    tx_conf = os.path.join(conf.paths.projectroot,
                           ".tx", 'config')

    with open(tx_conf, 'r') as f:
        files = [l.rsplit(' ', 1)[1].strip()
                 for l in f.readlines()
                 if l.startswith('source_file')]

    errs = 0
    for fn in files:
        fqfn = os.path.join(conf.paths.projectroot, fn)

        if not os.path.exists(fqfn):
            errs += 1
            logger.error(fqfn + " does not exist.")

    if errs != 0:
        logger.warning("{0} files configured that don't exist.")
    else:
        logger.info('all configured translation source files exist')

    return errs

# Task Generators


def pull_tasks(conf):
    resources = tx_resources(conf)
    tasks = []

    for page in resources:
        t = libgiza.task.Task(job=logged_command,
                              args=('pull', ' '.join(['tx', 'pull', '-l',
                                                      conf.runstate.language, '-r', page])),
                              target=True,
                              dependency=None,
                              description='pulling {0} from transifex client'.format(page))
        tasks.append(t)

    return tasks


def push_tasks(conf):
    resources = tx_resources(conf)
    tasks = []

    for page in resources:
        t = libgiza.task.Task(job=logged_command,
                              args=('pull', ' '.join(['tx', 'pull', '-l',
                                                      conf.runstate.language, '-r', page])),
                              target=True,
                              dependency=None,
                              description='pulling {0} from transifex client'.format(page))
        tasks.append(t)

    return tasks


def update_langauge(conf):
    logger.info('updating translation artifacts for commit.')

    project_name = conf.project.title.lower().split()
    if conf.project.edition is not None and conf.project.edition != conf.project.name:
        project_name.append(conf.project.edition)

    project_name = '-'.join(project_name)

    logger.info('starting translation upload with sphinx-intl')

    flogger = FileLogger(logger)

    sphinx_intl.commands.update(
        locale_dir=conf.paths.locale,
        pot_dir=os.path.join(conf.paths.projectroot,
                             conf.paths.locale, 'pot'),
        out=flogger)

    logger.info('sphinx-intl: updated pot directory')


def update_transifex(conf):
    logger.info('updating transifex artifacts. Long running.')

    project_name = conf.project.title.lower().split()
    if conf.project.edition is not None and conf.project.edition != conf.project.name:
        project_name.append(conf.project.edition)

    project_name = '-'.join(project_name)

    flogger = FileLogger(logger)

    sphinx_intl.commands.update_txconfig_resources(
        transifex_project_name=project_name,
        locale_dir=conf.paths.locale,
        pot_dir=os.path.join(conf.paths.projectroot,
                             conf.paths.locale, 'pot'),
        out=flogger)

    logger.info('sphinx-intl: updated pot directory')


# Commands

@argh.named('check')
@argh.expects_obj
def check_orphaned(args):
    conf = fetch_config(args)

    check_for_orphaned_tx_files(conf)


@argh.arg('--edition', '-e')
@argh.arg('--language', '-l')
@argh.named('update-transifex')
@argh.expects_obj
def update_translations_transifex(args):
    """Updates translation and uploads them to the transifex service."""

    conf = fetch_config(args)

    update_transifex(conf)
    check_for_orphaned_tx_files(conf)


@argh.arg('--edition', '-e')
@argh.arg('--language', '-l')
@argh.named('update')
@argh.expects_obj
def update_translations(args):
    """Updates translations in the locale directory.."""

    conf = fetch_config(args)

    update_langauge(conf)


@argh.named('pull')
@argh.expects_obj
def pull_translations(args):
    conf = fetch_config(args)

    with libgiza.app.BuildApp.new(pool_type=conf.runstate.runner,
                                  pool_size=conf.runstate.pool_size,
                                  force=conf.runstate.force).context() as app:
        app.extend_queue(pull_tasks(conf))


@argh.arg('--edition', '-e')
@argh.arg('--language', '-l')
@argh.named('push')
@argh.expects_obj
def push_translations(args):
    """Sync translations with the transifex service."""

    conf = fetch_config(args)

    with libgiza.app.BuildApp.new(pool_type=conf.runstate.runner,
                                  pool_size=conf.runstate.pool_size,
                                  force=conf.runstate.force).context() as app:
        app.extend_queue(push_tasks(conf))
        update_transifex(conf)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helpers and wrappers for performing common maintenance operations on a
repository with git.
"""

import logging
import os
import subprocess

import argh

from libgiza.app import BuildApp
from libgiza.git import GitRepo
from giza.config.helper import fetch_config
from giza.operations.build_env import fix_build_env_tasks, get_existing_builders

logger = logging.getLogger('giza.operations.git')


@argh.arg('--patch', '-p', nargs='*', dest='git_objects')
@argh.arg('--branch', '-b', nargs='*', dest='git_branch')
@argh.arg('--signoff', '-s', default=False, action='store_true', dest='git_sign_patch')
@argh.named('am')
@argh.expects_obj
def apply_patch(args):
    c = fetch_config(args)

    g = GitRepo(c.paths.projectroot)

    if c.runstate.git_branch is None:
        c.runstate.git_branch = [g.current_branch()]

    for branch in c.runstate.git_branch:
        with g.branch(branch):
            g.am(patches=c.runstate.git_objects,
                 repo='/'.join(['https://github.com', c.git.remote.upstream]),
                 sign=c.runstate.git_sign_patch)


@argh.arg('--branch', '-b', nargs="*", dest='git_branch')
@argh.named('update')
@argh.expects_obj
def pull_rebase(args):
    c = fetch_config(args)

    g = GitRepo(c.paths.projectroot)

    if c.runstate.git_branch is None:
        c.runstate.git_branch = [g.current_branch()]

    for branch in c.runstate.git_branch:
        with g.branch(branch):
            g.update()
            logger.info('updated: ' + branch)


@argh.arg('--branch', '-b', nargs="*", dest='git_branch')
@argh.arg('--commits', '-c', nargs='*', dest='git_objects')
@argh.named('cp')
@argh.expects_obj
def cherry_pick(args):
    c = fetch_config(args)

    g = GitRepo(c.paths.projectroot)

    if c.runstate.git_branch is None:
        c.runstate.git_branch = [g.current_branch()]

    for branch in c.runstate.git_branch:
        with g.branch(branch):
            g.cherry_pick(c.runstate.git_objects)


@argh.arg('--branch', '-b', default=None, dest='git_branch')
@argh.expects_obj
def merge(args):
    c = fetch_config(args)

    g = GitRepo(c.paths.projectroot)

    from_branch = g.current_branch()
    branch_name = str(id(c.runstate.git_branch))

    g.checkout_branch(branch_name, c.runstate.git_branch)

    try:
        g.checkout(branch_name)
        g.rebase(from_branch)
        g.checkout(from_branch)
        g.merge(c.runstate.git_branch)
        logger.info('rebased and merged {0} into {1}'.format(c.runstate.git_branch, from_branch))
    except Exception as e:
        logger.warning('error attempting to merge branch: ' + c.runstate.git_branch)
        logger.error(e)
    finally:
        if g.current_branch != from_branch:
            g.checkout(from_branch)

        g.remove_branch(branch_name, force=False)


@argh.expects_obj
@argh.named("setup-branches")
def setup_branches(args):
    conf = fetch_config(args)

    g = GitRepo(conf.paths.projectroot)

    if 'upstream' in g.remotes():
        remote = 'upstream'
    else:
        remote = 'origin'

    for pbranch in conf.git.branches.published:
        if g.branch_exists(pbranch):
            continue
        else:
            tracking_branch = '/'.join([remote, pbranch])
            g.create_branch(pbranch, tracking=tracking_branch)
            logger.info('created branch "{0}" tracking "{1}"'.format(pbranch, tracking_branch))


@argh.expects_obj
@argh.named("create-branch")
@argh.arg('git_branch')
def create_branch(args):
    """
    Takes a single branch name and (if necessary) creates a new branch. Then,
    populates the ``build/<branch>`` directory for the new branch using either
    the parent branch or ``master``. Safe to run multiple times (after a rebase)
    to update the build cache from master.

    Also calls :method:`~giza.operations.build_env.fix_build_environment()` to
    tweak the new build output to update hashes and on-disk copies of the
    environment to prevent unnecessary full-rebuilds from sphinx.
    """

    conf = fetch_config(args)

    g = GitRepo(conf.paths.projectroot)

    branch = conf.runstate.git_branch
    base_branch = g.current_branch()

    if base_branch == branch:
        base_branch = 'master'
        logger.warning('seeding build data for branch "{0}" from "master"'.format(branch))

    branch_builddir = os.path.join(conf.paths.projectroot,
                                   conf.paths.output, branch)

    base_builddir = os.path.join(conf.paths.projectroot,
                                 conf.paths.output, base_branch)

    if g.branch_exists(branch):
        logger.info('checking out branch "{0}"'.format(branch))
    else:
        logger.info('creating and checking out a branch named "{0}"'.format(branch))

    g.checkout_branch(branch)

    cmd = "rsync -r --times --checksum {0}/ {1}".format(base_builddir, branch_builddir)
    logger.info('seeding build directory for "{0}" from "{1}"'.format(branch, base_branch))

    try:
        subprocess.check_call(args=cmd.split())
        logger.info('branch creation complete.')
    except subprocess.CalledProcessError:
        logger.error(cmd)

    # get a new config here for the new branch
    conf = fetch_config(args)
    builders = get_existing_builders(conf)

    with BuildApp.new(pool_type='process',
                      pool_size=conf.runstate.pool_size,
                      force=conf.runstate.force).context() as app:
        app.exted_queue(fix_build_env_tasks(builders, conf))



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import shlex
import subprocess

import libgiza.task

import giza.content.images.views
import giza.tools.files
from giza.config.sphinx_config import resolve_builder_path

logger = logging.getLogger('giza.content.images')


def generate_image(build_type, dpi, width, target, source):
    import wand.image
    import wand.api
    import wand.color

    with wand.image.Image(filename=source, resolution=dpi) as image:
        if image.width != width:
            image.transform(resize=str(width))

        if build_type == 'png':
            alpha = 0.0
        elif build_type == 'eps':
            alpha = 1.0
        else:
            raise TypeError(build_type + " is not supported")

        try:
            image.transparent_color(color=wand.color.Color('white'), alpha=alpha)
        except:
            pass

        with open(target, 'wb') as out:
            out.write(image.make_blob(build_type))

    logger.info('wrote: ' + target)


def generate_image_inkscape(build_type, dpi, width, target, source):
    inkscape = None

    for path in ('/usr/bin/inkscape', '/usr/local/bin/inkscape',
                 '/Applications/Inkscape.app/Contents/Resources/bin/inkscape'):
        if os.path.exists(path):
            inkscape = path
            break

    if inkscape is None:
        logger.error("dependency INKSCAPE not installed. not building images.")
        return

    if build_type == 'png':
        cmd = '{cmd} -z -d {dpi} -w {width} -y 0.0 -e {target} {source}'
    elif build_type == 'eps':
        cmd = '{cmd} -z -d {dpi} -w {width} -y 1.0 -E {target} {source}'

    cmd = cmd.format(cmd=inkscape, dpi=dpi, width=width, target=target, source=source)
    logger.debug(cmd)
    with open(os.devnull, 'w') as null:
        r = subprocess.call(shlex.split(cmd), stdout=null, stderr=null)

    if r == 0:
        logger.info('wrote: ' + target)
    else:
        logger.warning('error generating image: ' + target)
        logger.error(cmd)


def image_tasks(conf, sconf):
    tasks = []

    deps = conf.system.files.get_configs('images')
    deps.append(os.path.abspath(__file__))

    if 'images' not in conf.system.files.data:
        return []

    giza.tools.files.safe_create_directory(os.path.join(conf.paths.projectroot,
                                                        conf.paths.branch_images))
    for image in conf.system.files.data.images:

        if not os.path.isfile(image.source_core):
            logger.error('"{0}" does not exist'.format(image.source_core))
            continue

        description = "generating rst include file {0} for {1}".format(image.rst_file,
                                                                       image.source_core)
        t = libgiza.task.Task(job=giza.content.images.views.generate_image_pages,
                              args=(image, conf),  # as kwargs
                              target=image.rst_file,
                              dependency=deps,
                              description=description)
        tasks.append(t)

        if conf.runstate.fast is True:
            continue

        for output in image.outputs:
            description = 'generating image file {0} from {1}'.format(output.output,
                                                                      image.source_core)
            t = libgiza.task.Task(job=generate_image_inkscape,
                                  args=(output.build_type, output.dpi, output.width,
                                        output.output, image.source_file),
                                  target=output.output,
                                  dependency=image.source_core,
                                  description=description)
            tasks.append(t)

            if output.type == 'target':
                image_output = os.path.join(conf.paths.projectroot,
                                            conf.paths.branch_output,
                                            resolve_builder_path(sconf.builder,
                                                                 conf.project.edition,
                                                                 None, conf),
                                            '_images',
                                            ''.join([image.name, '-', output.tag,
                                                     '.', output.build_type]))

                description = 'copying fullsize image file {0} from {1}'.format(image_output,
                                                                                output.output)

                t.add_finalizer(libgiza.task.Task(job=giza.tools.files.copy_if_needed,
                                                  args=(output.output, image_output),
                                                  description=description,
                                                  target=image_output,
                                                  dependency=None))

    logger.info('registered {0} image generation tasks'.format(len(tasks)))

    return tasks


def image_clean(conf):
    if 'images' not in conf.system.files.data:
        logger.info('no images to clean')
        return []

    tasks = []
    for image in conf.system.files.data.images:
        for output in image.outputs:
            t = libgiza.task.Task(job=giza.tools.files.verbose_remove,
                                  args=output.output,
                                  target=True,
                                  dependency=None,
                                  description="removing img file")
            tasks.append(t)

    return tasks



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Given a deploy target, we need to compile the rsync commands to:

- rsync the specified source directory to the specified target directory
- rsync all static files. (unless non-master and .htaccess)
- create the rsync command.

"""

import logging
import os.path
import subprocess
import shlex

logger = logging.getLogger('giza.deploy')


class Deploy(object):

    def __init__(self, conf):
        self.conf = conf
        self.name = None
        self.remote_path = None
        self.local_path = None
        self.delete = False
        self.recursive = True
        self.env = None
        self.hosts = None
        self.static_files = []
        self.branched = False

    def load(self, pspec):
        if 'target' in pspec:
            self.name = pspec['target']
        elif 'name' in pspec:
            self.name = pspec['name']

        self.remote_path = pspec['paths']['remote']
        self.local_path = pspec['paths']['local']

        if 'delete' in pspec['options']:
            self.delete = True

        if 'recursive' in pspec['options']:
            self.recursive = True

        if 'branched' in pspec['options']:
            self.branched = True

        self.env = pspec['env']

        self.deploy_env = getattr(self.conf.deploy, self.env)

        self.hosts = self.deploy_env.hosts

        if 'static' in pspec['paths']:
            self.static_files.extend(pspec['paths']['static'])

    def _base_cmd(self):
        base_cmd = ['rsync', '-cqltz']

        if self.delete is True:
            base_cmd.append('--delete')

        if self.recursive is True:
            base_cmd.append('--recursive')

        if 'args' in self.deploy_env:
            base_cmd.extend(self.deploy_env.args)

        return base_cmd

    def deploy_commands(self):
        base = self._base_cmd()

        for host in self.hosts:
            if self.branched is True:
                yield base + [os.path.join(self.conf.paths.output,
                                           self.local_path,
                                           self.conf.git.branches.current),
                              host + ':' + self.remote_path]
            else:
                yield base + [os.path.join(self.conf.paths.output, self.local_path) + '/',
                              host + ':' + self.remote_path]

            for fn in self.static_files:
                if self.conf.git.branches.current != 'master' and fn == '.htaccess':
                    logger.debug('skipping .htaccess files from non-master branch')
                    continue
                else:
                    yield base + [os.path.join(self.conf.paths.output, self.local_path, fn),
                                  host + ':' + self.remote_path]

    def run(self):
        map(deploy_target, self.deploy_commands())


def deploy_target(cmd):
    with open(os.devnull, 'w') as f:
        logger.info(cmd)
        r = subprocess.call(shlex.split(cmd), stderr=f, stdout=f)

    if r == 0:
        return 0
    elif r == 23:
        logger.warning('permissions error on remote end, possibly timestamp related.')
    elif r == 12:
        logger.warning('connection closed by remote host. rsync operation failed.')
    else:
        logger.error('"rsync" returned code {1}'.format(cmd, r))

    return r



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os

import libgiza.inheritance
import libgiza.config

import giza.content.helper

import giza.tools.files


class InheritanceReference(libgiza.inheritance.InheritanceReference):

    @property
    def file(self):
        return self.state['file']

    @file.setter
    def file(self, value):
        fns = [os.path.join(self.conf.paths.projectroot,
                            self.conf.paths.branch_includes, value),
               os.path.join(self.conf.paths.projectroot,
                            self.conf.paths.branch_source, value),
               os.path.join(self.conf.paths.projectroot, value)]

        for fn in fns:
            if os.path.exists(fn):
                self.state['file'] = fn
                break

        if 'file' not in self.state:
            raise TypeError('file named {0} does not exist'.format(value))


class InheritableContentBase(libgiza.inheritance.InheritableContentBase):
    _reference_type = InheritanceReference

    def get_default_replacement(self):
        if 'replacement' in self.conf.system.files.data:
            base = self.conf.system.files.data.replacement
            if isinstance(base, dict):
                base = copy.deepcopy(base)
            else:
                base = copy.deepcopy(base.dict())
        else:
            base = {}

        return base


class DataContentBase(libgiza.inheritance.DataContentBase):
    edition_check = staticmethod(giza.content.helper.edition_check)
    content_class = InheritableContentBase


class DataCache(libgiza.inheritance.DataCache):
    content_class = DataContentBase

    def create_output_dir(self):
        dirname = self.conf.system.content.get(self.content_type).output_dir
        if (self.content_type is not None and
                len(self) > 0 and
                not os.path.isdir(dirname)):

            giza.tools.files.safe_create_directory(dirname)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path
import sys

from giza.inheritance import InheritableContentBase
from giza.content.steps.models import HeadingMixin

logger = logging.getLogger('giza.content.extract.models')

if sys.version_info >= (3, 0):
    basestring = str


class ExtractData(HeadingMixin, InheritableContentBase):
    _default_level = 2

    @property
    def only(self):
        return self.state['only']

    @only.setter
    def only(self, value):
        if isinstance(value, basestring):
            self.state['only'] = value
        else:
            raise TypeError

    @property
    def append(self):
        return self._get_file('append')

    @append.setter
    def append(self, value):
        self._set_file(value, 'append')

    @property
    def prepend(self):
        return self._get_file('prepend')

    @prepend.setter
    def prepend(self, value):
        self._set_file(value, 'prepend')

    @property
    def style(self):
        return self.state['style']

    @style.setter
    def style(self, value):
        self.state['style'] = value

    @property
    def target(self):
        return os.path.join(self.conf.system.content.extracts.output_dir, self.ref) + '.rst'

    @property
    def target_project_path(self):
        offset = len(os.path.join(self.conf.paths.projectroot, self.conf.paths.branch_source))

        return self.target[offset:]

    # The actual implementation for prepend/append
    def _get_file(self, kind):
        if kind in self.state:
            return self.state[kind]
        else:
            return []

    def _set_file(self, value, kind):
        if not isinstance(value, list):
            value = [value]

        paths = []
        for fn in value:
            if fn.startswith(os.path.sep):
                fn = fn[1:]
            for path in [os.path.join(self.conf.paths.projectroot,
                                      self.conf.paths.branch_source, fn),
                         os.path.join(self.conf.paths.projectroot,
                                      self.conf.paths.branch_includes, fn),
                         os.path.join(self.conf.paths.projectroot, fn),
                         os.path.abspath(fn)]:
                if os.path.isfile(path):
                    paths.append(path)
                    break

        self.state[kind] = paths
        if len(value) > 0 and len(paths) < 0:
            logger.error('cannot {0} to non existing file "{1}", skipping.'.format(kind, value))



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Creates symbolic links in the build output based on definitions in the
(typically) "build/conf/integration.yaml" file in the ``base.links`` key.
"""

import os.path
import logging

from giza.tools.files import create_link

logger = logging.getLogger('giza.content.links')


def _link_path(path, conf):
    return os.path.join(conf.paths.projectroot,
                        conf.paths.public,
                        path)


def get_top_level_links(links, conf):
    ret = []

    def process_target_list(lst):
        for name, target in lst.items():
            if target == '{{current_branch}}':
                target = conf.git.branches.current

            yield (_link_path(name, conf), target)

    if isinstance(links, list):
        for link in links:
            ret.extend(process_target_list(link))
    else:
        ret.extend(process_target_list(links))

    return ret


def get_public_links(conf):
    iconf = conf.system.files.data.integration

    try:
        return get_top_level_links(iconf['base']['links'], conf)
    except KeyError:
        return []


def create_manual_symlink(conf):
    public_links = get_public_links(conf)

    for name, target in public_links:
        logger.info('creating link to "{0}", named "{1}"'.format(target, name))
        create_link(target, name)



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path

from libgiza.task import Task

from giza.content.tocs.inheritance import TocDataCache
from giza.content.tocs.views import render_toctree, render_dfn_list, render_toc_table
from giza.config.content import new_content_type

logger = logging.getLogger('giza.content.tocs.tasks')


def register_toc(conf):
    definition = new_content_type(name='toc',
                                  task_generator=toc_tasks,
                                  conf=conf,
                                  prefixes=['toc', 'ref-toc', 'ref-spec'])

    conf.system.content.add(name='toc', definition=definition)


def write_toc_tree_output(fn, toc_items, is_ref):
    content = render_toctree(toc_items, is_ref)
    content.write(fn)
    logger.info("wrote toctree to: " + fn)


def write_dfn_list_output(fn, toc_items):
    content = render_dfn_list(toc_items)
    content.write(fn)
    logger.info("wrote toc dfnlist to: " + fn)


def write_toc_table(fn, toc_items):
    content = render_toc_table(toc_items)
    content.write(fn)
    logger.info("wrote toc table to: " + fn)


def toc_tasks(conf):
    tocs = TocDataCache(conf.system.content.toc.sources, conf)
    tocs.create_output_dir()

    tasks = []
    for dep_fn, toc_data in tocs.file_iter():
        deps = [dep_fn]
        if 'ref-toc-' in dep_fn:
            base_offset = 8
            is_ref = True
        elif 'ref-spec-' in dep_fn:
            base_offset = 9
            is_ref = True
        else:
            base_offset = 4
            is_ref = False

        fn_basename = os.path.basename(dep_fn)[base_offset:].replace('yaml', 'rst')

        toc_items = toc_data.ordered_items()

        if toc_data.is_spec() is False:
            out_fn = os.path.join(conf.system.content.toc.output_dir, fn_basename)

            t = Task(job=write_toc_tree_output,
                     args=(out_fn, toc_items, is_ref),
                     target=out_fn,
                     dependency=dep_fn,
                     description="writing toctree to '{0}'".format(out_fn))
            tasks.append(t)
        else:
            deps.extend(toc_data.spec_deps())

        if 'ref-toc' in dep_fn:
            out_fn = os.path.join(conf.system.content.toc.output_dir, 'table-' + fn_basename)

            reft = Task(job=write_toc_table,
                        args=(out_fn, toc_items),
                        target=out_fn,
                        dependency=deps,
                        description="write ref toc table to '{0}'".format(out_fn))
            tasks.append(reft)
        elif 'ref-spec' in dep_fn:
            out_fn = os.path.join(conf.system.content.toc.output_dir, 'table-spec-' + fn_basename)

            refspec = Task(job=write_toc_table,
                           args=(out_fn, toc_items),
                           target=out_fn,
                           dependency=deps,
                           description="write ref spec table to '{0}'".format(out_fn))
            tasks.append(refspec)
        else:
            out_fn = os.path.join(conf.system.content.toc.output_dir, 'dfn-list-' + fn_basename)
            dt = Task(job=write_dfn_list_output,
                      args=(out_fn, toc_items),
                      target=out_fn,
                      dependency=deps,
                      description="write definition list toc to '{0}'".format(out_fn))
            tasks.append(dt)

    logger.info('added tasks for {0} toc generation tasks'.format(len(tasks)))

    return tasks



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys

from giza.inheritance import InheritableContentBase, InheritanceReference

logger = logging.getLogger('giza.content.options.models')

if sys.version_info >= (3, 0):
    basestring = str


class OptionData(InheritableContentBase):
    _option_registry = ['pre', 'post', 'final', 'ref', 'content', 'edition',
                        'description', 'name', 'args', 'aliases', 'default',
                        'type', 'filename']

    @property
    def source(self):
        if 'source' in self.state:
            return self.state['source']
        else:
            return None

    @source.setter
    def source(self, value):
        value['ref'] = (value['program'], value['name'])
        del value['program']
        del value['name']

        self.state['source'] = InheritanceReference(value, self.conf)

    inherit = source

    @property
    def program(self):
        return self.state['program']

    @program.setter
    def program(self, value):
        if isinstance(value, basestring):
            self.state['program'] = value
            self.state['ref'] = value
        else:
            raise TypeError

    @property
    def optional(self):
        if 'optional' in self.state:
            return self.state['optional']
        else:
            return True

    @optional.setter
    def optional(self, value):
        if isinstance(value, bool):
            self.state['optional'] = value
        else:
            raise TypeError

    @property
    def command(self):
        if 'command' in self.state:
            return self.state['command']

    @command.setter
    def command(self, value):
        if isinstance(value, basestring):
            self.state['command'] = value
        else:
            raise TypeError('command name must be a string')

    @property
    def filename(self):
        if 'filename' in self.state:
            return self.state['filename']

    @filename.setter
    def filename(self, value):
        if isinstance(value, basestring):
            self.state['filename'] = value
        else:
            raise TypeError('filename must be a string')

    @property
    def directive(self):
        if 'directive' in self.state:
            return self.state['directive']
        else:
            return 'object'

    @directive.setter
    def directive(self, value):
        if (value in ('option', 'commandoption', 'data', 'setting',
                      'method', 'function', 'class') or
                value.endswith('setting')):
            self.state['directive'] = value
        else:
            raise TypeError



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Glues the reading and processing (:mod:`giza.content.examples.inheritance`) of
the example data (:mod:`giza.content.examples.models`) to the generation of the
output format (:mod:`giza.content.examples.views`).
"""

import logging
import os
import shutil

from libgiza.task import Task

from giza.config.content import new_content_type
from giza.content.examples.inheritance import ExampleDataCache
from giza.content.examples.views import full_example

logger = logging.getLogger('giza.content.examples')


def register_examples(conf):
    content_dfn = new_content_type(name='examples',
                                   task_generator=example_tasks,
                                   conf=conf)

    conf.system.content.add(name='examples', definition=content_dfn)


def write_full_example(collection, examples, fn):
    content = full_example(collection, examples)
    content.write(fn)


def example_tasks(conf):
    d = ExampleDataCache(conf.system.content.examples.sources, conf)
    d.create_output_dir()

    tasks = []
    for fn, exmpf in d.file_iter():
        out_fn = os.path.join(conf.system.content.examples.output_dir,
                              conf.system.content.examples.get_basename(fn)) + '.rst'

        t = Task(job=write_full_example,
                 args=(exmpf.collection, exmpf.examples, out_fn),
                 description='generate an example for ' + fn,
                 target=out_fn,
                 dependency=fn)
        tasks.append(t)

    logger.info("added tasks for {0} example generation tasks".format(len(tasks)))
    return tasks


def example_clean(conf):
    return [Task(job=shutil.rmtree,
                 args=[conf.system.content.examples.output_dir],
                 target=True,
                 dependency=[conf.system.content.examples.output_dir],
                 description='removing {0}'.format(conf.system.content.examples.output_dir))]



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import tarfile
import logging
import hashlib

logger = logging.getLogger('giza.files')


class FileNotFoundError(Exception):
    pass


class InvalidFile(Exception):
    pass


class FileOperationError(Exception):
    pass


def safe_create_directory(path):
    try:
        os.makedirs(path)
        return True
    except OSError as e:
        if os.path.isdir(path):
            return None
        elif os.path.isfie(path):
            logger.error('"{0}" is a file not a directory: {1}'.format(path, e))
            raise e
        else:
            logger.error('encountered error creating directory: ' + path)
            raise e


def verbose_remove(path):
    if os.path.exists(path):
        logger.info('clean: removing {0}'.format(path))
        os.remove(path)


def tarball(name, path, newp=None, cdir=None):
    tarball_path = os.path.dirname(name)
    safe_create_directory(tarball_path)

    logger.debug('creating tarball: {0}'.format(name))
    with tarfile.open(name, 'w:gz') as t:
        if newp is not None:
            arcname = os.path.join(newp, os.path.basename(path))
        else:
            arcname = None

        if cdir is not None:
            path = os.path.join(cdir, path)

        logger.debug('tarball internal path {0}'.format(path))

        t.add(name=path, arcname=arcname)

    logger.info('created tarball: {0}'.format(name))


def symlink(name, target):
    if not os.path.islink(name):
        try:
            os.symlink(target, name)
        except AttributeError:
            from win32file import CreateSymbolicLink
            CreateSymbolicLink(name, target)
        except ImportError:
            logger.error("platform does not contain support for symlinks.")
            logger.info("Windows users need pywin32.")


def expand_tree(path, input_extension='yaml'):
    file_list = []

    for root, sub_folders, files in os.walk(path):
        for file in files:
            if file.startswith('.#'):
                continue
            elif file.endswith('swp'):
                continue
            else:
                f = os.path.join(root, file)
                if input_extension is not None:
                    if isinstance(input_extension, list):
                        if os.path.splitext(f)[1][1:] not in input_extension:
                            continue
                    else:
                        if not f.endswith(input_extension):
                            continue

                file_list.append(f)

    return file_list


def md5_file(file, block_size=2 ** 20):
    md5 = hashlib.md5()

    with open(file, 'rb') as f:
        for chunk in iter(lambda: f.read(128 * md5.block_size), b''):
            md5.update(chunk)

    return md5.hexdigest()


def copy_if_needed(source_file, target_file, name='build'):
    if os.path.isfile(source_file) is False or os.path.isdir(source_file):
        msg = "{0}: Input file '{1}' does not exist.".format(name, source_file)
        logger.critical(msg)
        raise FileOperationError(msg)
    elif os.path.isfile(target_file) is False:
        safe_create_directory(os.path.dirname(target_file))
        shutil.copyfile(source_file, target_file)

        if name is not None:
            logger.debug('{0}: created "{1}" which did not exist.'.format(name, target_file))
    else:
        if md5_file(source_file) == md5_file(target_file):
            if name is not None:
                logger.debug('{0}: "{1}" not changed.'.format(name, source_file))
        else:
            shutil.copyfile(source_file, target_file)

            if name is not None:
                m = '{0}: "{1}" changed. Updated: {2}'
                logger.debug(m.format(name, source_file, target_file))

    try:
        os.utime(target_file, None)
    except:
        pass


def copy_always(source_file, target_file, name='build'):
    if os.path.isfile(source_file) is False:
        msg = "{0}: Input file '{1}' does not exist.".format(name, source_file)
        logger.critical(msg)
        raise FileOperationError(msg)
    else:
        safe_create_directory(os.path.dirname(target_file))
        shutil.copyfile(source_file, target_file)

    logger.debug('{0}: copied {1} to {2}'.format(name, source_file, target_file))


def create_link(input_fn, output_fn):
    out_dirname = os.path.dirname(output_fn)

    if out_dirname != '':
        safe_create_directory(out_dirname)

    if os.path.islink(output_fn):
        os.remove(output_fn)
    elif os.path.isdir(output_fn):
        msg = "link: {1} exists and is a directory".format(output_fn)
        logger.critical(msg)
        raise FileOperationError(msg)
    elif os.path.exists(output_fn):
        msg = 'could not create a symlink at {1}.'.format('link', output_fn)
        logger.critical(msg)
        raise FileOperationError(msg)

    out_base = os.path.basename(output_fn)
    if out_base == "":
        msg = 'could not create a symlink at {1}.'.format('link', output_fn)
        logger.critical(msg)
        raise FileOperationError(msg)
    else:
        symlink(out_base, input_fn)
        os.rename(out_base, output_fn)
        m = '{0} created symbolic link pointing to "{1}" named "{2}"'
        logger.debug(m.format('symlink', input_fn, out_base))



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import copy
import logging

from giza.inheritance import InheritableContentBase, InheritanceReference

logger = logging.getLogger('giza.content.apiargs.models')

if sys.version_info >= (3, 0):
    basestring = str

field_type = {
    'param': 'Parameter',
    'field': 'Field',
    'arg': 'Argument',
    'option': 'Option',
    'flag': 'Flag'
}


class ApiArgData(InheritableContentBase):
    _option_registry = [
        'interface',  # TODO validate because limited possibilities
        'operation',
        'pre', 'post'
    ]

    @property
    def arg_name(self):
        return self.state['arg_name']

    @property
    def arg_name_rendered(self):
        if 'arg_name' not in self.state:
            return field_type['arg']
        else:
            if self.arg_name in field_type:
                return field_type[self.arg_name]
            else:
                return field_type['arg']

    @arg_name.setter
    def arg_name(self, value):
        if value in field_type:
            self.state['arg_name'] = value
        else:
            logger.error('incorrect argname format: ' + str(value))
            self.state['arg_name'] = value
            # raise TypeError

    @property
    def description(self):
        output = []

        for element in ("pre", "description", "post"):
            if element in self.state:
                output.append(self.state[element])
                output.append("\n")
        if self.optional is True:
            output[0] = "Optional. " + output[0]
        return "\n".join(output)

    @description.setter
    def description(self, value):
        if isinstance(value, list):
            value = '\n'.join(value)

        if not isinstance(value, basestring):
            raise TypeError

        if value.startswith('Optional.'):
            value = value[8:]
            self.optional = True

        self.state['description'] = value.strip()

    @property
    def type(self):
        if 'type' not in self.state:
            return ''
        else:
            return self.state['type']

    @type.setter
    def type(self, value):
        if 'type' not in self.state:
            self.state['type'] = []

        if isinstance(value, list):
            self.state['type'].extend(value)
        else:
            self.state['type'].append(value)

    def type_for_field_output(self):
        return ', '.join(self.type)

    def type_for_table_output(self):
        if len(self.type) == 0:
            return ''
        elif len(self.type) == 1:
            return self.type[0]
        elif len(self.type) == 2:
            return ' or '.join(self.type)
        else:
            tmp = copy.copy(self.type)
            tmp[-1] = ' or ' + tmp[-1]
            return ', '.join(tmp)

    @property
    def source(self):
        if 'source' in self.state:
            return self.state['source']
        else:
            return None

    @source.setter
    def source(self, value):
        if 'name' not in self.state and 'ref' not in self.state:
            try:
                self.state['name'] = value['name']
            except KeyError:
                self.state['ref'] = value['ref']

        self.state['source'] = InheritanceReference(value, self.conf)

    inherit = source

    @property
    def ref(self):
        if 'name' in self.state:
            return self.state['name']
        elif 'ref' in self.state:
            return self.state['ref']
        else:
            raise ValueError

    @ref.setter
    def ref(self, value):
        self.state['ref'] = value

    @property
    def name(self):
        if 'name' in self.state:
            return self.state['name']
        elif 'ref' in self.state:
            return self.state['ref']
        else:
            raise TypeError

    @name.setter
    def name(self, value):
        if isinstance(value, basestring):
            self.state['name'] = value
            self.state['ref'] = value
        else:
            raise TypeError

    @property
    def position(self):
        if not hasattr(self, '_position'):
            self._position = None

        return self._position

    @position.setter
    def position(self, value):
        if not hasattr(self, '_position'):
            self._position = None

        if isinstance(value, (int, float, complex)):
            self._position = int(value)
            self.state['number'] = self._position
        else:
            raise TypeError

    number = position

    @property
    def optional(self):
        if 'optional' in self.state:
            return self.state['optional']
        else:
            return False

    @optional.setter
    def optional(self, value):
        if value is True:
            self.state['optional'] = True
        else:
            self.state['optional'] = False



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import libgiza.task

from giza.tools.files import copy_always, copy_if_needed

logger = logging.getLogger('giza.transformation')


class ProcessingError(Exception):
    pass


def decode_lines_from_file(fn):
    with open(fn, 'r') as f:
        return [line.decode('utf-8').rstrip() for line in f.readlines()]


def encode_lines_to_file(fn, lines):
    with open(fn, 'w') as f:
        f.write('\n'.join(lines).encode('utf-8'))
        f.write('\n')


def munge_page(fn, regex, out_fn=None,  tag='build'):
    if out_fn is None:
        out_fn = fn

    page_lines = [munge_content(ln, regex) for ln in decode_lines_from_file(fn)
                  if ln is not None]

    if len(page_lines) > 0:
        encode_lines_to_file(out_fn, page_lines)
    else:
        logger.warning('{0}: did not write {1}'.format(tag, out_fn))


def munge_content(content, regex):
    if isinstance(regex, list):
        for cregex, subst in regex:
            content = cregex.sub(subst, content)
        return content
    else:
        return regex[0].sub(regex[1], content)


def truncate_file(fn, start_after=None, end_before=None):
    if start_after is not None and end_before is not None:
        if type(start_after) != type(end_before):
            raise TypeError('start-after and end-before types must match')

    with open(fn, 'r') as f:
        source_lines = f.readlines()

    should_find_line_num = False

    if isinstance(start_after, int):
        start_idx = start_after
    else:
        # start_after is none or some string -- if string, find line num
        start_idx = 0
        if start_after is not None:
            should_find_line_num = True

    if isinstance(end_before, int):
        end_idx = end_before - 1
    else:
        # end_before is none or some string -- if string, find line num
        end_idx = len(source_lines) - 1
        if end_before is not None:
            should_find_line_num = True

    # should_find_line_num is True if:
    #  - start_after = string and end_before = string
    #  - start_after = string and end_before is None
    #  - start_after is None and end_before is string

    if should_find_line_num is True:
        for idx, ln in enumerate(source_lines):
            if start_after is not None and start_after in ln:
                start_idx = idx + 1
                if end_before is None:
                    break

            if end_before is not None and end_before in ln:
                end_idx = idx
                break

    with open(fn, 'w') as f:
        f.writelines(source_lines[start_idx:end_idx])


def append_to_file(fn, text):
    with open(fn, 'a') as f:
        f.write('\n')
        f.write(text)


def prepend_to_file(fn, text):
    with open(fn, 'r') as f:
        body = f.readlines()

    with open(fn, 'w') as f:
        f.write(text)
        f.writelines(body)


def process_page_task(fn, output_fn, regex, builder='processor', copy='always'):
    return libgiza.task.Task(job=_process_page,
                             args=(fn, output_fn, regex, copy, builder),
                             target=output_fn,
                             dependency=None,
                             description="modify page: ({0}, {1})".format(fn, output_fn))


def process_page(fn, output_fn, regex, app, builder='processor', copy='always'):
    t = app.add('task')
    t.job = _process_page
    t.args = [fn, output_fn, regex, copy, builder]
    t.target = output_fn
    t.depenency = None
    t.description = "modify page"

    logger.debug('added tasks to process file: {0}'.format(fn))


def _process_page(fn, output_fn, regex, copy, builder):
    tmp_fn = fn + '~'

    munge_page(fn=fn, out_fn=tmp_fn, regex=regex)

    cp_args = dict(source_file=tmp_fn,
                   target_file=output_fn,
                   name=builder)

    if copy == 'always':
        copy_always(**cp_args)
    else:
        copy_if_needed(**cp_args)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

from giza.tools.files import verbose_remove
from giza.content.options.inheritance import OptionDataCache
from giza.content.options.views import render_options
from giza.config.content import new_content_type
from libgiza.task import Task

logger = logging.getLogger('giza.content.options.tasks')


def register_options(conf):
    content_dfn = new_content_type(name='option',
                                   task_generator=option_tasks,
                                   conf=conf)

    conf.system.content.add(name='options', definition=content_dfn)


def write_options(option, fn, conf):
    content = render_options(option, conf)
    content.write(fn)
    logger.info('wrote options file: ' + fn)


def option_tasks(conf):
    o = OptionDataCache(conf.system.content.options.sources, conf)
    o.create_output_dir()

    tasks = []
    for dep_fn, option in o.content_iter():
        program = option.program.replace(' ', '-')

        if option.has_field('filename'):
            option_name = option.filename
        else:
            option_name = option.name

        output_fn = os.path.join(conf.system.content.options.fn_prefix,
                                 ''.join((option.directive, '-', program,
                                          '-', option_name + '.rst')))

        t = Task(job=write_options,
                 args=(option, output_fn, conf),
                 description='generating option file "{0}" from "{1}"'.format(output_fn, dep_fn),
                 target=output_fn,
                 dependency=[dep_fn])
        tasks.append(t)

    logger.info("added tasks for {0} option generation tasks".format(len(tasks)))
    return tasks


def option_clean(conf):
    register_options(conf)

    tasks = []
    for fn in conf.system.options.sources:
        t = Task(job=verbose_remove,
                 args=[fn],
                 target=fn,
                 dependency=None,
                 description='removing {0}'.format(fn))

        tasks.append(t)

    return tasks



# 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rstcloth.rstcloth import RstCloth


def get_replacements(conf):
    if "replacement" in conf.system.files.data:
        mapping = conf.system.files.data.replacement
    else:
        return []

    r = RstCloth()

    try:
        if conf.version.release != "Upcoming":
            mapping['release-string'] = "-- {0} Release".format(conf.version.release)
        else:
            mapping['release-string'] = "\ "
    except:
        pass

    for k, v in mapping.items():
        r.replacement(k, v)

    return r.data



import os
import logging
import argh
import giza.content.changelog.views

logger = logging.getLogger('giza.jeerah.client')


@argh.expects_obj
@argh.arg("changelog_version", default=None)
@argh.named("cl")
def main(args):
    pass



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os.path
import logging

import libgiza.config
import sphinx.builders
import yaml

logger = logging.getLogger('giza.config.sphinx_config')

# Ingestion and Rendering


def is_legacy_sconf(conf):
    if 'v' not in conf or conf['v'] == 0:
        return True
    else:
        return False


def get_sconf_base(conf):
    sconf_path = os.path.join(conf.paths.projectroot, conf.paths.builddata, 'sphinx.yaml')
    with open(sconf_path, 'r') as f:
        sconf = yaml.safe_load(f)

    return sconf


def render_sconf(edition, builder, language, conf):
    sconf_base = get_sconf_base(conf)

    sconf = SphinxConfig(conf, sconf_base)
    sconf.register(builder, language, edition)

    return sconf

# Helpers


def resolve_builder_path(builder, edition, language, conf):
    dirname = builder

    if edition is not None and edition != conf.project.name:
        dirname = '-'.join((dirname, edition))

    if language is not None and language != 'en':
        dirname = '-'.join((dirname, language))

    return dirname


def avalible_sphinx_builders():
    builders = sphinx.builders.BUILTIN_BUILDERS.keys()
    builders.append('slides')
    builders.append('publish')

    return builders

# New-Style Config Object


class SphinxConfig(libgiza.config.RecursiveConfigurationBase):
    _option_registry = ['languages']

    def __init__(self, conf, input_obj=None):
        # register the global config, but use our ingestion
        # super(SphinxConfig, self).__init__(None, conf)

        self._state = {}
        self._raw = {}
        self.conf = conf
        self.ingest(input_obj)

    def ingest(self, input_obj=None):
        if input_obj is None:
            input_obj = get_sconf_base(self.conf)

        if is_legacy_sconf(input_obj):
            self._raw = render_sphinx_config(input_obj)
        else:
            self._raw = input_obj

    def register(self, builder, language, edition):
        self.language = language
        self.builder = builder
        self.edition = edition

        lookup_opts = []

        if edition is None:
            lookup_opts.append(self.builder)
        else:
            lookup_opts.append('-'.join((self.builder, self.edition)))

        base = {}
        for opt in lookup_opts:
            if opt in self._raw:
                base = self._raw[opt]
                break

        for i in ['excluded', 'tags', 'languages']:
            if i in base:
                setattr(self, i, base[i])

        m = 'registered language, builder, and edition options: ({0}, {1}, {2})'
        logger.debug(m.format(language, builder, edition))

    @property
    def name(self):
        if 'name' in self.state:
            return self.state['name']
        else:
            return self.builder

    @name.setter
    def name(self, value):
        self.state['name'] = value

    @property
    def build_output(self):
        if 'build_output' not in self.state:
            self.build_output = None

        return self.state['build_output']

    @build_output.setter
    def build_output(self, value):
        if value is None:
            path = resolve_builder_path(self.builder, self.edition, self.language, self.conf)
            self.state['build_output'] = path
        else:
            self.state['build_output'] = path

    @property
    def fq_build_output(self):
        if 'build_output' not in self.state:
            self.build_output = None

        return os.path.join(self.conf.paths.projectroot,
                            self.conf.paths.branch_output,
                            self.build_output)

    @property
    def language(self):
        if 'language' not in self.state:
            return 'en'
        else:
            return self.state['language']

    @language.setter
    def language(self, value):
        self.state['language'] = value

    @property
    def edition(self):
        if 'edition' in self.state:
            return self.state['edition']
        else:
            if ('edition' in self.conf.project and
                    self.conf.project.edition != self.conf.project.name):

                return self.conf.project.edition
            else:
                return ''

    @edition.setter
    def edition(self, value):
        self.state['edition'] = value

    @property
    def builder(self):
        return self.state['builder']

    @property
    def excluded(self):
        if 'excluded' not in self.state:
            return []
        else:
            return self.state['excluded']

    @property
    def tags(self):
        return self.state['tags']

    @property
    def languages(self):
        return self.state['languages']

    @builder.setter
    def builder(self, value):
        if value in avalible_sphinx_builders():
            self.state['builder'] = value
        else:
            raise TypeError('{0} is not a valid sphinx builder'.format(value))

    @excluded.setter
    def excluded(self, value):
        if not isinstance(value, list):
            logger.error('excluded files must be a list.')
            raise TypeError
        else:
            if 'excluded' in self.state:
                self.state['excluded'].extend(value)
            else:
                self.state['excluded'] = value

    @tags.setter
    def tags(self, value):
        if not isinstance(value, list):
            logger.error('builder tags must be a list.')
            raise TypeError
        else:
            if 'tags' in self.state:
                self.state['tags'].extend(value)
            else:
                self.state['tags'] = value

    @languages.setter
    def languages(self, value):
        if not isinstance(value, list):
            logger.error('languages must be a list.')
            raise TypeError
        else:
            self.state['languages'] = value


# Rendering for Legacy Config


def resolve_legacy_sphinx_config(sconf_base, edition, builder, language):
    sconf_base = render_sphinx_config(sconf_base)

    if edition is not None:
        builder = '-'.join([builder, edition])

    sconf = sconf_base[builder]

    sconf['edition'] = edition
    if 'builder' not in sconf:
        sconf['builder'] = builder

    if language is not None:
        sconf['language'] = language

    return sconf


def render_sphinx_config(conf):
    computed = {}

    def resolver(v, conf, computed):
        while 'inherit' in v:
            if v['inherit'] in computed:
                base = copy.deepcopy(computed[v['inherit']])
            else:
                base = copy.deepcopy(conf[v['inherit']])

            if 'excluded' in v and 'excluded' in base:
                if isinstance(v['excluded'], list):
                    base['excluded'].extend(v['excluded'])
                    del v['excluded']

            del v['inherit']
            base.update(v)
            v = base

        return v

    to_compute = []

    for k, v in conf.items():
        v = resolver(v, conf, computed)
        computed[k] = v

        if 'languages' in v:
            to_compute.append((k, v))

    for k, v in to_compute:
        if k in ['prerequisites', 'generated-source',
                 'sphinx-builders'] or 'base' in k:
            continue

        if 'languages' in v:
            for lang in v['languages']:
                computed['-'.join([k, lang])] = resolver({'inherit': k,
                                                          'language': lang},
                                                         conf, computed)

    for i in computed.keys():
        if i in ['prerequisites', 'generated-source',
                 'sphinx-builders'] or 'base' in i:
            del computed[i]

    return computed



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Post-process all output produced by JSON to add a "text" field to these
documents that omits the XML data injected into this format by default so that
search tools can use this data to index content. Also generates a file with a
list of paths in the output.
"""

import json
import logging
import os
import re
import subprocess

import libgiza.task

from giza.tools.files import expand_tree, copy_if_needed, safe_create_directory
from giza.tools.transformation import munge_content

logger = logging.getLogger('giza.content.post.json_output')

# Process Sphinx Json Output


def json_output(conf):
    list_file = os.path.join(conf.paths.branch_output, 'json-file-list')
    public_list_file = os.path.join(conf.paths.projectroot,
                                    conf.paths.public_site_output,
                                    'json', '.file_list')

    cmd = ('rsync --recursive --times --delete '
           '--exclude="*pickle" --exclude=".buildinfo" --exclude="*fjson" '
           '{src} {dst}')

    json_dst = os.path.join(conf.paths.projectroot, conf.paths.public_site_output, 'json')
    safe_create_directory(json_dst)

    builder = 'json'
    if 'edition' in conf.project and conf.project.edition != conf.project.name:
        builder += '-' + conf.project.edition

    cmd_str = cmd.format(src=os.path.join(conf.paths.projectroot,
                                          conf.paths.branch_output, builder) + '/',
                         dst=json_dst)

    try:
        subprocess.check_call(cmd_str.split())
        copy_if_needed(list_file, public_list_file)
        logger.info('deployed json files to local staging.')
    except subprocess.CalledProcessError:
        logger.error('error migrating json artifacts to local staging')


def json_output_tasks(conf):
    regexes = [
        (re.compile(r'<a class=\"headerlink\"'), '<a'),
        (re.compile(r'<[^>]*>'), ''),
        (re.compile(r'&#8220;'), '"'),
        (re.compile(r'&#8221;'), '"'),
        (re.compile(r'&#8216;'), "'"),
        (re.compile(r'&#8217;'), "'"),
        (re.compile(r'&#\d{4};'), ''),
        (re.compile(r'&nbsp;'), ''),
        (re.compile(r'&gt;'), '>'),
        (re.compile(r'&lt;'), '<')
    ]

    outputs = []

    tasks = []

    for fn in expand_tree('source', 'txt'):
        # path = build/<branch>/json/<filename>

        if 'edition' in conf.project and conf.project.edition != conf.project.name:
            path = os.path.join(conf.paths.branch_output,
                                'json-' + conf.project.edition,
                                os.path.splitext(fn.split(os.path.sep, 1)[1])[0])

        else:
            path = os.path.join(conf.paths.branch_output,
                                'json', os.path.splitext(fn.split(os.path.sep, 1)[1])[0])

        fjson = path + '.fjson'
        jsonf = path + '.json'

        task = libgiza.task.Task(job=process_json_file,
                                 args=(fjson, jsonf, regexes, conf),
                                 target=jsonf,
                                 dependency=fjson,
                                 description="processing json file {}".format(json))
        tasks.append(task)
        outputs.append(jsonf)

    list_file = os.path.join(conf.paths.branch_output, 'json-file-list')
    tasks.append(libgiza.task.Task(job=generate_list_file,
                                   args=(outputs, list_file, conf),
                                   target=list_file,
                                   dependency=None,
                                   description="generating list of json files"))

    transfer = libgiza.task.Task(job=json_output,
                                 args=[conf],
                                 target=True,
                                 dependency=None,
                                 description='transfer json output to public directory')

    return tasks, transfer


def process_json_file(input_fn, output_fn, regexes, conf=None):
    if os.path.isfile(input_fn) is False:
        return False

    with open(input_fn, 'r') as f:
        document = f.read()

    doc = json.loads(document)

    if 'body' in doc:
        text = doc['body'].encode('ascii', 'ignore')
        text = munge_content(text, regexes)

        doc['text'] = ' '.join(text.split('\n')).strip()

    if 'title' in doc:
        title = doc['title'].encode('ascii', 'ignore')
        title = munge_content(title, regexes)

        doc['title'] = title

    url = get_site_url(conf)
    url.extend(input_fn.rsplit('.', 1)[0].split(os.path.sep)[3:])
    doc['url'] = '/'.join(url) + '/'

    with open(output_fn, 'w') as f:
        f.write(json.dumps(doc))

    return True


def generate_list_file(outputs, path, conf):
    dirname = os.path.dirname(path)
    safe_create_directory(dirname)

    url = get_site_url(conf)
    url.append('json')
    url = '/'.join(url)

    with open(path, 'w') as f:
        for fn in outputs:
            if os.path.isfile(fn) is True:
                line = '/'.join([url, fn.split('/', 3)[3:][0]])
                f.write(line)
                f.write('\n')

    logger.info('rebuilt inventory of json output.')


def get_site_url(conf):
    url = [conf.project.url]

    if conf.project.basepath not in ('', None):
        url.append(conf.project.basepath)

    if conf.project.branched is True and 'editions' in conf.project:
        url.append(conf.git.branches.current)

    return url



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import logging

import giza.content.changelog.views
import giza.tools.files

from giza.config.content import new_content_type
from libgiza.task import Task

logger = logging.getLogger('giza.content.changelog.tasks')


def register_changelogs(conf):
    content_dfn = new_content_type(name='changelogs',
                                   task_generator=changelog_tasks,
                                   conf=conf)

    conf.system.content.add(name='changelogs', definition=content_dfn)


def get_major_version_groupings(versions):
    major_versions = {}
    for v in versions:
        parts = [int(i) for i in v.split('.')]
        mver = '.'.join([str(s) for s in parts[0:2]])
        if mver not in major_versions:
            major_versions[mver] = [parts]
        else:
            major_versions[mver].append(parts)

    for v in major_versions.values():
        v.sort(reverse=True)

    return major_versions


def changelog_tasks(conf):
    tasks = []

    if 'jira' not in conf.system.files.data:
        logger.debug('changelog generation is not configured.')
        return []

    dirname = os.path.join(conf.paths.projectroot, conf.paths.includes, 'changelogs')

    giza.tools.files.safe_create_directory(os.path.join(dirname, 'releases'))
    jira_config = os.path.join(conf.paths.projectroot, conf.paths.builddata, 'jira.yaml')
    major_versions = get_major_version_groupings(conf.system.files.data.jira.site.versions)

    # If no version listed in jira.yaml, just return; Should be same
    # as if jira is not configured  in conf.system.files.data
    # Also, log explicit message stating that 0 changelog tasks added.
    if not major_versions:
        logger.warning('changelog version is not configured in jira.yaml.')
        logger.info('added {0} changelog tasks.'.format(len(tasks)))
        return []

    # don't generate changelog content except on the most recent published
    # branch (i.e. master, typically.).
    if conf.git.branches.current != conf.git.branches.published[0]:
        logger.error('you must generate changelogs on the master branch and them backport them to another branch.')
        logger.info('added {0} changelog tasks'.format(len(tasks)))
        return tasks

    # only generate changelogs if there are credentials, even though we don't
    # really need an auth'ed connection, want to avoid making un-authed builds too long.
    if not os.path.exists(os.path.expanduser(conf.system.files.data.jira.site.credentials)):
        logger.warning('jira credentials are not configured for your user. not generating changelog tasks')
        logger.info('added {0} changelog tasks'.format(len(tasks)))
        return tasks

    # bump mtime of all existing files to avoid regenerating files that already committed files
    # exist.
    changelog_releases_dir = os.path.join(conf.paths.projectroot, conf.paths.includes, 'changelogs', 'releases')

    for fn in os.listdir(changelog_releases_dir):
        os.utime(os.path.join(changelog_releases_dir, fn), None)

    # add tasks for generating intermediate files for each major version. we do
    # this on all branches, and publishers need to backport the config changes.
    for version, releases in major_versions.items():
        fn = os.path.join(dirname, version + '.rst')
        t = Task(job=giza.content.changelog.views.render_intermediate_files,
                 args=(fn, version, releases, conf),
                 target=fn,
                 dependency=[jira_config])
        tasks.append(t)

    # create a task for each version defined. should never regenerate existing files.
    for version in conf.system.files.data.jira.site.versions:
        fn = os.path.join(conf.paths.projectroot, conf.paths.includes, 'changelogs', 'releases', version + '.rst')
        t = Task(job=giza.content.changelog.views.get_changelog_content,
                 args=(fn, version, conf),
                 dependency=[jira_config],
                 target=fn)
        tasks.append(t)

    logger.info('added {0} changelog tasks.'.format(len(tasks)))
    return tasks



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Responsible for migrating content from the ``source`` repository to a directory
in ``build/<branch>/source`` (per-edition). These "proxy-source" directories
make it possible to:

- build while editing the source and changing sources,

- limit content generation in the source tree itself,

- avoid changing the ``mtime`` of files during branch changes to facilitate
  more incremental builds,

- have different versions of the source tree for different editions of the
  content (i.e. by redacting files or modifying the source,)

At the center of this operation is an ``rsync`` operation that uses check-summing
rather than timestampping to compare source and destination files.
"""

import os.path
import logging
import shutil
import shlex
import subprocess

import libgiza.task

from giza.tools.files import InvalidFile, safe_create_directory

logger = logging.getLogger('giza.content.source')

# Transfer Source Files


def transfer_source(conf, sconf):
    target = os.path.join(conf.paths.projectroot, conf.paths.branch_source)

    dir_exists = safe_create_directory(target)

    # this operation is just for messaging the above operation, and error'ing
    # appropriately.
    if dir_exists is True:
        logger.info('created directory for sphinx build: {0}'.format(target))
    elif not os.path.isdir(target):
        msg = '"{0}" exists and is not a directory'.format(target)
        logger.error(msg)
        raise InvalidFile(msg)

    source_dir = os.path.join(conf.paths.projectroot, conf.paths.source)
    image_dir = os.path.join(conf.paths.images[len(conf.paths.source) + 1:])

    exclusions = [os.path.join('includes', 'table'),
                  os.path.join('includes', 'generated'),
                  image_dir + os.path.sep + "*.png",
                  image_dir + os.path.sep + "*.rst"]

    prefix_len = len(os.path.join(conf.paths.projectroot, conf.paths.branch_source)) + 1

    exclusions.extend([o for o in conf.system.content.output_directories(prefix_len)
                       if o != "includes/changelogs"])

    # we don't want rsync to delete directories that hold generated content in
    # the target so we can have more incremental builds.
    exclusions = "--exclude=" + ' --exclude='.join(exclusions)

    cmd = 'rsync --links --checksum --recursive {2} --delete {0}/ {1}'
    cmd = cmd.format(source_dir, target, exclusions)

    try:
        subprocess.check_call(shlex.split(cmd))
    except subprocess.CalledProcessError as e:
        logger.error('source transfer rsync had error: ' + str(e.returncode))
        logger.info(cmd)

    # remove files from the source tree specified in the sphinx config for this
    # build.
    source_exclusion(conf, sconf)
    os.utime(target, None)

    logger.info('prepared and migrated source for sphinx build in {0}'.format(target))


def source_exclusion(conf, sconf):
    ct = 0
    if len(sconf.excluded) == 0:
        return

    for fn in sconf.excluded:
        fqfn = os.path.join(conf.paths.projectroot, conf.paths.branch_source, fn[1:])
        if os.path.isdir(fqfn):
            shutil.rmtree(fqfn)
            ct += 1
        elif os.path.isfile(fqfn):
            os.remove(fqfn)
            ct += 1
            logger.debug('redacted {0}'.format(fqfn))
        else:
            logger.warning('cannot redact non-existing file: ' + fqfn)

    logger.info('redacted {0} files'.format(ct))

# Transfer Images

# transfer all ``.eps`` images to the latex build directory because to generate
# "offset" resources, we declare the image in raw latex and Sphinx cannot migrate these images.


def transfer_images(conf, sconf):
    image_dir = os.path.join(conf.paths.projectroot, conf.paths.branch_images)
    if not os.path.isdir(image_dir):
        return False
    elif sconf.builder == 'latex':

        if 'edition' in sconf and sconf.edition is not None:
            builder_dir = '-'.join((sconf.builder, sconf.edition))
        else:
            builder_dir = sconf.builder

        builder_dir = os.path.join(conf.paths.projectroot, conf.paths.branch_output, builder_dir)
        safe_create_directory(builder_dir)

        cmd = ('rsync -am '
               '--include="*.png" --include="*.jpg" --include="*.eps" '
               '--exclude="*" {0}/ {1}')
        cmd = cmd.format(image_dir, builder_dir)

        with open(os.devnull, 'w') as null:
            for img_cmd in (shlex.split(cmd.replace('images', 'figures')),
                            shlex.split(cmd)):

                subprocess.call(img_cmd, stdout=null, stderr=null)

        logger.info('migrated images for latex build')

# Task Creators


def latex_image_transfer_tasks(conf, sconf):
    description = 'transferring images to build directory to {0}'.format(conf.paths.branch_source)

    return [libgiza.task.Task(job=transfer_images,
                              args=(conf, sconf),
                              target=True,
                              description=description)]


def source_tasks(conf, sconf):
    description = 'migrating source to {0}'.format(conf.paths.branch_source)

    return [libgiza.task.Task(job=transfer_source,
                              args=(conf, sconf),
                              target=os.path.join(conf.paths.branch_source),
                              description=description)]



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A function that creates tasks that migrate ``.po`` files generated by Sphinx's
``gettext`` builder to the ``locale`` directory in the source directory to use
as a basis for translation.
"""

import os.path
import logging

import libgiza.task

from giza.tools.files import expand_tree, copy_if_needed
from giza.config.sphinx_config import resolve_builder_path

logger = logging.getLogger('giza.content.post.gettext')

# Gettext Processing


def gettext_tasks(conf):
    locale_dirs = os.path.join(conf.paths.projectroot,
                               conf.paths.locale, 'pot')

    builder_name = resolve_builder_path('gettext', conf.project.edition, None, conf)

    branch_output = os.path.join(conf.paths.projectroot,
                                 conf.paths.branch_output,
                                 builder_name)

    path_offset = len(branch_output) + 1

    tasks = []
    for fn in expand_tree(branch_output, None):
        target = os.path.join(locale_dirs, fn[path_offset:])
        source = fn

        t = libgiza.task.Task(job=copy_if_needed,
                              args=(source, target, None),
                              target=target,
                              dependency=source,
                              description="migrating po file {0} if needed".format(fn))
        tasks.append(t)

    logger.info("if you added files to the corpus since your last gettext build,"
                "please run the gettext build a second time for complete finalization.")

    return tasks



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging


class ColorFormatter(logging.Formatter):
    """Logging formatter using VT100 terminal color codes."""
    VT100 = {
        'red': '31',
        'yellow': '33',
        'bright': '1'
    }

    COLORS = {
        logging.WARNING: ('yellow',),
        logging.ERROR: ('red', 'bright'),
        logging.CRITICAL: ('red', 'bright'),
    }

    def __init__(self, fmt='%(levelname)s:%(name)s:%(message)s', color=True):
        """Create a pretty-printing logging formatter. If color is True, use
           VT100 color codes to colorify output."""
        logging.Formatter.__init__(self, fmt=fmt)
        self.color = color

    def format(self, record):
        # If the "lean" option is given, don't include any additional information
        msg = logging.Formatter.format(self, record)
        if getattr(record, 'lean', False):
            msg = record.msg

        if not self.color:
            return msg

        composite = []
        for option in self.COLORS.get(record.levelno, ()):
            composite.append('{0}'.format(self.VT100[option]))

        return '\x1b[{0}m{1}\x1b[0m'.format(';'.join(composite), msg)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Generates per-build "buildinfo" artifacts that access to build-time data,
including the release.txt file that has the hash that reflects the version of
the source reflected in a build, as well as content useable in the includes
directory so that you can reference the commit in the documentation text.
"""

import logging
import os

import libgiza.task
from rstcloth.rstcloth import RstCloth

logger = logging.getLogger('giza.hash')

# Rendering


def generate_hash_file(fn, conf):
    r = RstCloth()

    if os.path.exists(fn):
        with open(fn, 'r') as f:
            existing = f.read()
    else:
        existing = []

    commit = conf.git.commit
    r.directive('|commit| replace', '``{0}``'.format(commit))

    try:
        if r.data == existing[:-1]:
            logger.info('no new commit(s), not updating {0} ({1})'.format(fn, commit[:10]))
            return True
    except TypeError:
        logger.warning('problem generating {0}, continuing'.format(fn))
        if os.path.exists(fn):
            os.utime(fn, None)
        else:
            with open(fn, 'a'):
                os.utime(fn, None)
    else:
        r.write(fn)
        logger.info('regenerated {0} with new commit hash: {1}'.format(fn, commit[:10]))


def generate_release_file(release_fn, conf):
    release_root = os.path.dirname(release_fn)
    if not os.path.exists(release_root):
        os.makedirs(release_root)

    with open(release_fn, 'w') as f:
        f.write(conf.git.commit)

    logger.info('generated "{0}" with current release hash.'.format(release_fn))

# Worker


def hash_tasks(conf):
    hash_fn = os.path.join(conf.paths.projectroot,
                           conf.paths.branch_includes,
                           'hash.rst')

    release_fn = os.path.join(conf.paths.projectroot,
                              conf.paths.public_site_output,
                              'release.txt')

    return [libgiza.task.Task(job=generate_hash_file,
                              args=(hash_fn, conf),
                              target=hash_fn,
                              dependency=None,
                              description='creating hash file: {0}'.format(hash_fn)),
            libgiza.task.Task(job=generate_release_file,
                              args=(release_fn, conf),
                              target=hash_fn,
                              dependency=None,
                              description="creating release filename: {0}".format(release_fn))]



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os.path
import logging

from libgiza.config import ConfigurationBase, RecursiveConfigurationBase

logger = logging.getLogger('giza.config.jeerah')

if sys.version_info >= (3, 0):
    basestring = str


class JeerahConfig(RecursiveConfigurationBase):
    @property
    def buckets(self):
        if 'buckets' in self.state:
            return self.state['buckets']
        else:
            return {}

    @property
    def site(self):
        return self.state['site']

    @site.setter
    def site(self, value):
        if isinstance(value, JeerahSiteConfig):
            self.state['site'] = value
        else:
            self.state['site'] = JeerahSiteConfig(value)

    @property
    def changelog(self):
        return self.state['changelog']

    @changelog.setter
    def changelog(self, value):
        # changelog canfig can either be specified in the the project, or we can
        # source it from the global config, if a filename is defined.
        cconfig = ChangelogConfiguration()

        if isinstance(value, dict):
            if 'source' in value:
                value = os.path.join(self.conf.paths.projectroot,
                                     self.conf.paths.global_config,
                                     value['source'])

            cconfig.ingest(value)

            self.state['changelog'] = cconfig
        else:
            raise TypeError('invalid changelog')


class ChangelogConfiguration(ConfigurationBase):
    @property
    def ordering(self):
        return self.state['ordering']

    @ordering.setter
    def ordering(self, value):
        if isinstance(value, list):
            invalid = []
            for item in value:
                if not isinstance(item, basestring):
                    invalid.append(item)

            if len(invalid) > 0:
                raise TypeError('{0} item(s) are not strings.'.format(', '.join(invalid)))
            else:
                self.state['ordering'] = value

        else:
            raise TypeError('changelog ordering must be list.')

    @property
    def nesting(self):
        return self.state['nesting']

    @nesting.setter
    def nesting(self, value):
        self._set_dict2list_mapping('nesting', value)

    @property
    def groups(self):
        return self.state['groups']

    @groups.setter
    def groups(self, value):
        self._set_dict2list_mapping('groups', value)

    def _set_dict2list_mapping(self, name, value):
        if isinstance(value, dict):
            invalid = []

            for k, v in value.items():
                if isinstance(v, list):
                    continue
                else:
                    invalid.append(k)

            if len(invalid) > 0:
                raise TypeError('item(s) {} are not valid {} definitions.'.format(
                    ', '.join(invalid), name))
            else:
                self.state[name] = value
        else:
            raise TypeError('invalid {0} definition.'.format(name))


class JeerahSiteConfig(ConfigurationBase):
    _option_registry = ['url']

    @property
    def credentials(self):
        return self.state['credentials']

    @credentials.setter
    def credentials(self, value):
        value = os.path.expanduser(value)
        self.state['credentials'] = value

    @property
    def projects(self):
        return self.state['projects']

    @projects.setter
    def projects(self, value):
        if isinstance(value, list):
            for item in value:
                if not isinstance(item, basestring):
                    raise TypeError('jira project {0} is not a string'.format(value))

            self.state['projects'] = value
        else:
            raise TypeError('jira projects must be a list: {0}'.format(value))

    @property
    def versions(self):
        return self.state['versions']

    @versions.setter
    def versions(self, value):
        if isinstance(value, list):
            for item in value:
                if not isinstance(item, basestring):
                    raise TypeError('jira version {0} is not a string'.format(value))

                if len(item.split('.')) != 3:
                    raise TypeError('{0} is an invalid version'.format(item))

            self.state['versions'] = value
        else:
            raise TypeError('jira versions must be a list: {0}'.format(value))



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Entry points that send locally built content to web servers.
"""

import logging

import argh
import onetimepass as otp

from giza.config.helper import fetch_config, new_credentials_config
from libgiza.app import BuildApp
from giza.deploy import Deploy, deploy_target
from giza.operations.sphinx_cmds import sphinx_publication

logger = logging.getLogger('giza.operations.deploy')


@argh.arg('--target', '-t', nargs='*', dest='push_targets')
@argh.arg('--dry-run', '-d', action='store_true', dest='dry_run')
@argh.named('deploy')
@argh.expects_obj
def main(args):
    """
    Uploads all build artifacts to the production environment. Does not build or
    render artifacts.
    """

    c = fetch_config(args)
    app = BuildApp.new(pool_type=c.runstate.runner,
                       pool_size=c.runstate.pool_size,
                       force=c.runstate.force)

    deploy_tasks(c, app)

    if c.runstate.dry_run is False:
        app.run()


@argh.arg('--deploy', '-d', nargs='*', dest='push_targets')
@argh.arg('--edition', '-e', nargs='*', dest='editions_to_build')
@argh.arg('--language', '-l', nargs='*', dest='languages_to_build')
@argh.arg('--builder', '-b', nargs='*', default='html')
@argh.arg('--serial_sphinx', action='store_true')
@argh.named('push')
@argh.expects_obj
def publish_and_deploy(args):
    """
    Combines the work of ``giza sphinx`` and ``giza deploy``, to produce build
    artifacts and then upload those artifacts to the servers.
    """

    c = fetch_config(args)
    app = BuildApp.new(pool_type=c.runstate.runner,
                       pool_size=c.runstate.pool_size,
                       force=c.runstate.force)

    sphinx_ret = sphinx_publication(c, app)
    if sphinx_ret == 0 or c.runstate.force is True:
        deploy_tasks(c, app)

        if c.runstate.dry_run is False:
            app.run()
    else:
        logger.warning(sphinx_ret + ' sphinx build(s) failed, and build not forced. not deploying.')


def deploy_tasks(c, app):
    """
    Deploys the build. The logic for generating the rsync commands is
    in ``giza.deploy``, and the configuration data is typically in
    ``config/push``.

    This function glues the config with the rsync command creation and then
    runs the commands.
    """
    pconf = dict((item['target'], item) for item in c.system.files.data.push)

    for target in c.runstate.push_targets:
        d = Deploy(c)

        target_pconf = pconf[target]

        if target_pconf['env'] == 'publication':
            target_pconf['env'] = 'production'

        d.load(target_pconf)

        for cmd in d.deploy_commands():
            task = app.add('task')
            task.args = ' '.join(cmd)
            task.job = deploy_target

            if c.runstate.dry_run is True:
                logger.info('dry run: {0}'.format(' '.join(cmd)))

    logger.info('completed deploy for: {0}'.format(' '.join(c.runstate.push_targets)))


@argh.named('code')
@argh.expects_obj
def twofa_code(args):
    "Returns a 2 factor authentication code."

    creds = new_credentials_config()

    print(otp.get_totp(creds.corp.seed))



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Post-processes Sphinx's latex output and generate PDFs from these ``tex`` files.
"""

import logging
import os
import re
import subprocess
import shlex

import libgiza.task

from giza.content.helper import edition_check
from giza.tools.transformation import process_page_task
from giza.tools.files import create_link, copy_if_needed

logger = logging.getLogger('giza.content.post.latex')

# PDFs from Latex Produced by Sphinx


def _render_tex_into_pdf(fn, deployed_path, path, output_format="pdf"):
    """
    Runs ``pdflatex`` operations, can generate ``dvi`` and ``pdf``. Runs
    pdflatex multiple times to correctly index and cross reference the PDF.
    """

    inputs_path = ".:{0}:".format(path)
    os.environ['TEXINPUTS'] = inputs_path

    if output_format == 'dvi':
        cmd = 'pdflatex --output-format dvi --interaction batchmode --output-directory {0} {1}'
        pdflatex = cmd.format(path, fn)
    elif output_format == 'pdf':
        pdflatex = 'pdflatex --interaction batchmode --output-directory {0} {1}'.format(path, fn)
    else:
        logger.error('not rendering pdf because {0} is not an output format'.format(output_format))
        return

    base_fn = os.path.basename(fn)
    cmds = [pdflatex,
            "makeindex -s {0}/python.ist {0}/{1}.idx ".format(path, base_fn[:-4]),
            pdflatex,
            pdflatex]

    if output_format == 'dvi':
        cmds.append("dvipdf {0}.dvi".format(base_fn[:-4]))

    with open(os.devnull, 'w') as null:
        for idx, cmd in enumerate(cmds):
            ret = subprocess.call(args=shlex.split(cmd),
                                  cwd=path,
                                  stdout=null,
                                  stderr=null)
            if ret == 0:
                m = 'pdf completed rendering stage {0} of {1} successfully ({2}, {3}).'
                logger.info(m.format(idx + 1, len(cmds), base_fn, ret))
                continue
            else:
                if idx <= 1:
                    m = 'pdf build encountered error early on {0}, continuing cautiously.'
                    logger.warning(m.format(base_fn))
                    continue
                else:
                    m = 'pdf build encountered error running pdflatex, investigate {0}. terminating'
                    logger.error(m.format(base_fn))
                    logger.error(' '.join(['TEXINPUTS={0} '.format(inputs_path),
                                           cmd.replace('--interaction batchmode', '')]))
                    return False

    pdf_fn = os.path.splitext(fn)[0] + '.pdf'
    copy_if_needed(pdf_fn, deployed_path, 'pdf')


def pdf_tasks(sconf, conf):
    """Returns a list of Tasks() to generate all PDFs."""

    tasks = []
    target = sconf.builder
    if 'pdfs' not in conf.system.files.data:
        return []

    # a list of tuples in (compileRegex, substitution) format.
    tex_regexes = [
        (re.compile(r'(index|bfcode)\{(.*)--(.*)\}'),
         r'\1\{\2-\{-\}\3\}'),
        (re.compile(r'\\PYGZsq{}'), "'"),
        (re.compile(r'\\code\{/(?!.*{}/|etc|usr|data|var|srv|data|bin|dev|opt|proc|24|private)'),
         r'\code{' + conf.project.url + r'/' + conf.project.tag + r'/')]

    # the path that sphinx writes tex files to are are different for editions.
    if 'edition' in conf.project and conf.project.edition != conf.project.name:
        latex_dir = os.path.join(conf.paths.projectroot,
                                 conf.paths.branch_output,
                                 '-'.join((target, conf.project.edition)))
    else:
        latex_dir = os.path.join(conf.paths.projectroot, conf.paths.branch_output, target)

    deploy_path = os.path.join(conf.paths.projectroot, conf.paths.public_site_output)

    # special case operations on "offset pdfs", which use EPS images.
    if 'tags' in sconf and "offset" in sconf.tags:
        output_format = "dvi"
        sty_file = os.path.join(latex_dir, 'sphinx.sty')
        t = process_page_task(fn=sty_file,
                              output_fn=sty_file,
                              regex=(re.compile(r'\\usepackage\[pdftex\]\{graphicx\}'),
                                     r'\usepackage{graphicx}'),
                              builder='sphinx-latex',
                              copy='ifNeeded')
        tasks.append(t)
    else:
        output_format = "pdf"

    for i in conf.system.files.data.pdfs:
        if edition_check(i, conf) is False:
            continue

        # compatibility shim for new/old images
        i = i.dict()

        if len(i['tag']) == 0:
            tagged_name = i['output'][:-4]
        else:
            tagged_name = i['output'][:-4] + '-' + i['tag']

        deploy_fn = tagged_name + '-' + conf.git.branches.current + '.pdf'
        link_name = deploy_fn.replace('-' + conf.git.branches.current, '')

        i['source'] = os.path.join(latex_dir, i['output'])
        i['processed'] = os.path.join(latex_dir, tagged_name + '.tex')
        i['pdf'] = os.path.join(latex_dir, tagged_name + '.pdf')
        i['deployed'] = os.path.join(deploy_path, deploy_fn)
        i['link'] = os.path.join(deploy_path, link_name)
        i['path'] = latex_dir

        # add the processing task
        t = process_page_task(fn=i['source'],
                              output_fn=i['processed'],
                              regex=tex_regexes,
                              builder='tex-munge',
                              copy='ifNeeded')
        tasks.append(t)

        # add task for changing TEX to PDF. (this also copies the pdf to the deployed path).
        render_task = libgiza.task.Task(job=_render_tex_into_pdf,
                                        args=(i['processed'], i['deployed'],
                                              i['path'], output_format),
                                        target=i['pdf'],
                                        dependency=None)  # i['processed']
        t.finalizers.append(render_task)

        # if needed create links.
        if i['link'] != i['deployed']:
            link_task = libgiza.task.Task(job=create_link,
                                          args=(deploy_fn, i['link']),
                                          target=i['link'],
                                          dependency=i['deployed'])
            render_task.finalizers.append(link_task)

    return tasks



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path
import docutils.core
from rstcloth.rstcloth import RstCloth

logger = logging.getLogger('giza.content.images.views')


def generate_image_pages(image, conf):
    r = RstCloth()

    dir = image.dir
    name = image.name
    alt = image.alt
    output = image.outputs
    image = os.path.sep.join([dir, name])

    for img_output in output:
        width = str(img_output.width) + 'px'
        build_type = img_output.build_type

        r.newline()

        if 'tag' in img_output:
            tag = ''.join(['-', img_output.tag, '.', build_type])
        else:
            tag = '.' + build_type

        options = [('alt', alt), ('align', 'center'), ('figwidth', width)]

        if 'scale' in img_output:
            options.append(('scale', img_output.scale))
        if 'target' in img_output:
            options.append(('target', (img_output.target)))

        if img_output.type == 'target':
            continue
        elif img_output.type == 'print':
            r.directive('only', 'latex and not offset', wrap=False)
            r.newline()

            r.directive(name='figure',
                        arg='/images/{0}{1}'.format(name, tag),
                        fields=options,
                        indent=3)
        elif img_output.type == 'offset':
            tex_figure = [
                r'\begin{figure}[h!]',
                r'\centering',
                ''.join([r'\includegraphics[width=', width,
                         ']{', name, tag, '}']),
                r'\end{figure}'
            ]

            r.directive('only', 'latex and offset', wrap=False)
            r.newline()
            r.directive('raw', 'latex', content=tex_figure, indent=3)
        else:
            r.directive('only', 'website and slides', wrap=False)
            r.newline()
            r.directive(name='figure',
                        arg='/images/{0}{1}'.format(name, tag),
                        fields=options,
                        indent=3)

            r.newline()

            r.directive('only', 'website and html', wrap=False)
            r.newline()
            r.directive(name='figure',
                        arg='/images/{0}{1}'.format(name, tag),
                        fields=options,
                        indent=3)

            r.newline()

            if img_output.width > 740:
                options[2] = ('figwidth', '740px')

            r.directive('only', 'website and not (html or slides)', wrap=False)
            r.newline()
            img_str = ''.join(['<div class="figure align-center" style="max-width:{5};">',
                               '<img src="{0}/{1}/_images/{2}{3}" alt="{4}">', '</img>',
                               '{6}</div>'])
            alt_html = docutils.core.publish_parts(alt, writer_name='html')['body'].strip()
            r.directive(name='raw', arg='html',
                        content=img_str.format(conf.project.url,
                                               conf.git.branches.current, name, tag, alt,
                                               width, alt_html),
                        indent=3)

        r.newline()

    image_rst_file_path = os.path.join(conf.paths.projectroot, image + '.rst')
    r.write(image_rst_file_path)
    logger.info('generated include file {0}.rst'.format(image))



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
import os

import libgiza.config
import yaml

logger = logging.getLogger('giza.config.test')

if sys.version_info >= (3, 0):
    basestring = str


class TestConfig(libgiza.config.ConfigurationBase):
    def ingest(self, input_obj):
        if os.path.isfile(input_obj) is True:
            with open(input_obj, 'r') as f:
                input_obj = [d for d in yaml.safe_load_all(f)]
        elif isinstance(input_obj, list):
            pass
        else:
            raise RuntimeError('invalid test configuration')

        for doc in input_obj:
            self.add_project(doc)

    def add_project(self, project):
        if isinstance(project, TestProject):
            self.projects.append(project)
        else:
            self.projects.append(TestProject(project))

    @property
    def projects(self):
        if 'projects' not in self.state:
            self.projects = None
        return self.state['projects']

    @projects.setter
    def projects(self, value):
        if 'projects' not in self.state:
            self.state['projects'] = []

        if value is None:
            return
        elif isinstance(value, list):
            for proj in value:
                self.add_project(proj)
        else:
            self.add_projecct(value)

    @property
    def private_projects(self):
        for project in self.projects:
            if project.private is True:
                yield project

    @property
    def public_projects(self):
        for project in self.projects:
            if project.private is False:
                yield project


class TestProject(libgiza.config.ConfigurationBase):
    @property
    def project(self):
        return self.state['project']

    @project.setter
    def project(self, value):
        if isinstance(value, basestring):
            self.state['project'] = value
        else:
            raise TypeError(type(value), value)

    @property
    def uri(self):
        return self.state['uri']

    @uri.setter
    def uri(self, value):
        if not isinstance(value, basestring):
            raise TypeError(type(value), value)
        elif not value.endswith('.git'):
            raise TypeError('malformed uri: ' + value)

        self.state['uri'] = value

    @property
    def branches(self):
        if 'branches' not in self.state:
            return ['master']
        else:
            return self.state['branches']

    @branches.setter
    def branches(self, value):
        if isinstance(value, basestring):
            value = [value]

        if not isinstance(value, list):
            raise TypeError(type(value), value)
        else:
            for branch in value:
                if not isinstance(branch, basestring):
                    raise TypeError(type(value), value)

        self.state['branches'] = value

    @property
    def private(self):
        if 'private' in self.state:
            return self.state['private']
        else:
            return False

    @private.setter
    def private(self, value):
        if isinstance(value, bool):
            self.state['private'] = value
        else:
            self.state['private'] = bool(value)

    @property
    def root(self):
        if 'root' in self.state:
            return self.state['root']
        else:
            return None

    @root.setter
    def root(self, value):
        if isinstance(value, basestring):
            self.state['root'] = value
        else:
            raise TypeError(type(value), value)

    @property
    def operations(self):
        if 'operations' in self.state:
            return self.state['operations']
        else:
            return {}

    @operations.setter
    def operations(self, value):
        if not isinstance(value, dict):
            raise TypeError(type(value), value)

        ops = {}
        for key, op in value.items():
            ops[key] = []
            for cmd in op:
                if isinstance(cmd, basestring):
                    ops[key].append(cmd)
                elif isinstance(cmd, list):
                    ops[key].extend(cmd)

        self.state['operations'] = ops



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argh
import os
import logging
import subprocess
import shlex
import shutil

import libgiza.git
import libgiza.app
import libgiza.task

import giza.content.assets
import giza.config.helper
import giza.config.main
import giza.tools.files

logger = logging.getLogger('GIZA.OPERATIONS.TEST')


def setup_test_repo(path, project):
    if os.path.isdir(path):
        g = libgiza.git.GitRepo(path)
        if g.current_branch() != 'master':
            g.checkout_branch('master')
        g.pull(remote='origin', branch='master')
        logger.info('updated repository at: ' + path)
    else:
        g = libgiza.git.GitRepo(os.path.dirname(path))
        g.clone(project.uri, os.path.basename(path))
        logger.info('cloned new repository into: ' + path)

    return project


def get_test_config(args):
    try:
        conf = giza.config.helper.fetch_config(args)
    except RuntimeError:
        path = os.path.join('data', 'build_config.yaml')
        if not os.path.isfile(path):
            logger.warning('must run test from the docs-tools repo, or a giza project directory.')
            raise SystemExit(-1)
        else:
            args.conf_path = path

            conf = giza.config.main.Configuration()
            conf.ingest(args.conf_path)
            conf.runstate = args
            conf.paths.projectroot = os.getcwd()

    return conf


def change_branch(path, branch):
    g = libgiza.git.GitRepo(path)
    tracking = '/'.join(('origin', branch))

    if g.branch_exists(branch) is True:
        g.checkout(branch)
        g.pull(remote='origin', branch=branch)
        logger.info('checked out and updated {0} ({1}) in {2}'.format(branch, tracking, g.path))
    else:
        g.checkout_branch(branch, tracking=tracking)

        logger.info('checked out {0} ({1}) in {2}'.format(branch, tracking, g.path))


def run_test_op(cmd, dir):
    g = libgiza.git.GitRepo(dir)

    r = subprocess.call(args=shlex.split(cmd), cwd=dir)
    if r != 0:
        m = 'failure with {0}, in "{1}", ({2})'.format(cmd, dir, g.current_branch())
        logger.error(m)
        raise RuntimeError(m)
    else:
        logger.info('completed {0}, in "{1}", ({2})'.format(cmd, dir, g.current_branch()))
        return 0


def cleaner(path):
    rm_path = os.path.join(path, "build")
    if not os.path.exists(rm_path):
        logger.info('directory is clean: ' + rm_path)
    else:
        shutil.rmtree(os.path.join(path, "build"))
        logger.info("removed path: " + rm_path)


integration_targets = ('complete', 'minimal', 'cleanComplete', 'cleanMinimal')


@argh.arg('--branch', '-b', dest='_override_branch', nargs="*", default=None)
@argh.arg('--project', '-p', dest='_override_projects', nargs="*", default=None)
@argh.arg('--operation', '-o', dest='_test_op', default='complete', choices=integration_targets)
@argh.expects_obj
@argh.named('test')
def integration_main(args):
    conf = get_test_config(args)
    app = libgiza.app.BuildApp.new(pool_type=conf.runstate.runner,
                                   pool_size=conf.runstate.pool_size,
                                   force=conf.runstate.force)

    build_path = os.path.join(conf.paths.projectroot, conf.paths.output)

    giza.tools.files.safe_create_directory(build_path)

    for project in conf.test.projects:
        if (args._test_op not in project.operations and
                args._test_op.lower()[5:] not in project.operations):
            m = 'operation {0} not defined for project {1}'
            logger.error(m.format(args._test_op, project))
            continue

        if (args._override_projects is not None and
                project.project not in args._override_projects):
            continue

        if args._override_branch is not None:
            project.branches = args._override_branch

        if project.root is None:
            path = os.path.join(build_path, project.project)
            git_path = path
        else:
            path = os.path.join(build_path, project.project, project.root)
            git_path = os.path.join(build_path, project.project)

        task = app.add(libgiza.task.Task(job=setup_test_repo,
                                         args=(git_path, project),
                                         ignore=False))

        if args._test_op.startswith('clean'):
            task = task.add_finalizer(libgiza.task.Task(job=cleaner,
                                                        args=[path],
                                                        ignore=False))

        for branch in project.branches:
            task = task.add_finalizer(libgiza.task.Task(job=change_branch,
                                                        args=(path, branch),
                                                        ignore=False))

            if args._test_op.startswith('clean'):
                op_name = args._test_op.lower()[5:]
            else:
                op_name = args._test_op

            for op in project.operations[op_name]:
                task = task.add_finalizer(libgiza.task.Task(job=run_test_op,
                                                            args=(op, path),
                                                            ignore=False))

    app.run()



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from giza.inheritance import DataContentBase, DataCache
from giza.content.apiargs.models import ApiArgData

logger = logging.getLogger('giza.content.inheritance.apiargs')


class ApiArgFile(DataContentBase):
    content_class = ApiArgData

    def field_type(self):
        name = set()

        for content in self.content.values():
            name.add(content.arg_name_rendered)

        if len(name) > 1:
            logger.warning('too many field types returning one at random.')

        return name.pop()

    def has_type(self):
        for content in self.content.values():
            if 'type' in content:
                return True

        return False


class ApiArgDataCache(DataCache):
    content_class = ApiArgFile
    content_type = 'apiargs'



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Generate ``rst`` tables from ``yaml`` files in the ``source/includes`` directory
that begin with ``table-``. Uses the table builder capacity in the
:mod:`rstcloth.table` module.

Legacy implementation.
"""

import os.path
import logging

import libgiza.task

from rstcloth.table import TableBuilder, YamlTable, ListTable
from giza.tools.files import expand_tree, verbose_remove, safe_create_directory

logger = logging.getLogger('giza.content.table')

# Table Builder


def _get_table_output_name(fn):
    base, leaf = os.path.split(os.path.splitext(fn)[0])

    return os.path.join(base, 'table', leaf[6:]) + '.rst'


def _get_list_table_output_name(fn):
    base, leaf = os.path.split(os.path.splitext(fn)[0])

    return ''.join((os.path.join(base, 'table', leaf[6:]), '-list', '.rst'))


def make_parent_dirs(*paths):
    for path in paths:
        safe_create_directory(os.path.dirname(path))


def _generate_tables(source, target, list_target):
    table_data = YamlTable(source)

    make_parent_dirs(target, list_target)

    # if not table_data.format or table_data.format is None:
    #     build_all = True

    list_table = TableBuilder(ListTable(table_data))
    list_table.write(list_target)
    logger.debug('rebuilt rendered table %s', list_target)

    list_table.write(target)
    logger.debug('rebuilt rendered list table %s', target)

    # if build_all or table_data.format == 'list':
    #     list_table = TableBuilder(ListTable(table_data))
    #     list_table.write(list_target)
    #     print('[table]: rebuilt {0}'.format(list_target))
    # if build_all or table_data.format == 'rst':
    #     # this really ought to be RstTable, but there's a bug there.
    #     rst_table = TableBuilder(ListTable(table_data))
    #     rst_table.write(target)

    #     print('[table]: rebuilt {0} as (a list table)'.format(target))

    logger.info('rebuilt rendered table output for %s', source)


# Table Source Iterators

def table_sources(conf):
    path = os.path.join(conf.paths.projectroot, conf.paths.branch_includes)

    for source in expand_tree(path, 'yaml'):
        if os.path.basename(source).startswith('table'):
            yield source

# Table Tasks


def table_tasks(conf):
    tasks = []
    for source in table_sources(conf):
        target = _get_table_output_name(source)
        list_target = _get_list_table_output_name(source)

        description = 'generating tables: {}, {} from {}'.format(target, list_target, source)
        tasks.append(libgiza.task.Task(job=_generate_tables,
                                       args=(source, target, list_target),
                                       target=[target, list_target],
                                       dependency=source,
                                       description=description))

        logger.debug('adding table job to build: %s', target)

    return tasks


def table_clean(conf):
    tasks = []
    for source in table_sources(conf):
        tasks.extend([libgiza.task.Task(job=verbose_remove,
                                        args=[_get_table_output_name(source)]),
                      libgiza.task.Task(job=verbose_remove,
                                        args=[_get_list_table_output_name(source)])])

    return tasks



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from giza.inheritance import DataContentBase, DataCache
from giza.content.tocs.models import TocData

logger = logging.getLogger('giza.content.tocs.inheritance')


class TocFile(DataContentBase):
    content_class = TocData

    def ordered_items(self):
        return [
            self.fetch(ref)
            for ref in self.ordering
        ]

    def is_spec(self):
        for content_item in self.content.values():
            if content_item.source is None:
                continue
            else:
                return True

        return False

    def spec_deps(self):
        deps = []
        for content_item in self.content.values():
            if content_item.source is None:
                continue
            else:
                deps.append(content_item.source.file)

        return deps


class TocDataCache(DataCache):
    content_class = TocFile
    content_type = 'toc'



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from giza.inheritance import DataContentBase, DataCache
from giza.content.options.models import OptionData

logger = logging.getLogger('giza.content.options.inheritance')


class OptionFile(DataContentBase):
    content_class = OptionData

    def add(self, doc):
        content = self.content_class(doc, self.conf)

        ref = (content.program, content.name)

        if not content.is_resolved():
            content.resolve(self.data)

        self.content[ref] = content

        return content


class OptionDataCache(DataCache):
    content_class = OptionFile
    content_type = 'options'



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

from libgiza.task import Task

from giza.content.apiargs.inheritance import ApiArgDataCache
from giza.content.apiargs.views import render_apiargs
from giza.config.content import new_content_type

logger = logging.getLogger('giza.content.apiargs.tasks')


def register_apiargs(conf):
    content_def = new_content_type(name='apiargs',
                                   task_generator=apiarg_tasks,
                                   conf=conf)

    conf.system.content.add(name='apiargs', definition=content_def)


def write_apiargs(apiargs, fn):
    content = render_apiargs(apiargs)
    content.write(fn)
    logger.info('wrote apiarg table to: ' + fn)


def apiarg_tasks(conf):
    a = ApiArgDataCache(conf.system.content.apiargs.sources, conf)
    a.create_output_dir()

    tasks = []
    for dep_fn, apiargs in a.file_iter():
        basename = conf.system.content.steps.get_basename(dep_fn)[2:]
        out_fn = os.path.join(conf.system.content.apiargs.output_dir, basename) + '.rst'

        t = Task(job=write_apiargs,
                 args=(apiargs, out_fn),
                 target=out_fn,
                 dependency=dep_fn,
                 description="write apiarg table for: " + dep_fn)
        tasks.append(t)

    logger.info('added tasks for {0} apiarg table generation tasks'.format(len(tasks)))

    return tasks



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from libgiza.config import ConfigurationBase

logger = logging.getLogger('giza.config.helper')


class IntersphinxConfig(ConfigurationBase):
    _option_registry = ['name']

    @property
    def url(self):
        return self.state['url']

    @url.setter
    def url(self, value):
        if value.startswith('http'):
            self.state['url'] = value
        else:
            raise TypeError

    @property
    def path(self):
        return self.state['path']

    @path.setter
    def path(self, value):
        if value.endswith('inv'):
            self.state['path'] = value
        else:
            raise TypeError



# 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path
import re
import sys
import subprocess
import shutil

from giza.tools.transformation import munge_page
from giza.tools.files import create_link, copy_if_needed
from giza.content.post.singlehtml import get_single_html_dir

logger = logging.getLogger('giza.content.post.sites')

# Sphinx Post-Processing


def finalize_epub_build(builder, conf):
    epub_name = '-'.join(conf.project.title.lower().split())
    epub_branched_filename = epub_name + '-' + conf.git.branches.current + '.epub'
    epub_src_filename = epub_name + '.epub'

    copy_if_needed(source_file=os.path.join(conf.paths.projectroot,
                                            conf.paths.branch_output,
                                            builder, epub_src_filename),
                   target_file=os.path.join(conf.paths.projectroot,
                                            conf.paths.public_site_output,
                                            epub_branched_filename))
    create_link(input_fn=epub_branched_filename,
                output_fn=os.path.join(conf.paths.projectroot,
                                       conf.paths.public_site_output,
                                       epub_src_filename))


def error_pages(sconf, conf):
    builder = sconf.builder

    if 'errors' not in conf.system.files.data:
        return None
    else:
        sub = (re.compile(r'\.\./\.\./'), conf.project.url + r'/' + conf.project.tag + r'/')

        for idx, error in enumerate(conf.system.files.data.errors):
            page = os.path.join(conf.paths.projectroot,
                                conf.paths.branch_output, builder,
                                'meta', error, 'index.html')
            munge_page(fn=page, regex=sub, tag='error-pages')

        logger.info('error-pages: rendered {0} error pages'.format(idx))


def finalize_dirhtml_build(sconf, conf):
    builder = sconf.builder

    single_html_dir = get_single_html_dir(conf)
    search_page = os.path.join(conf.paths.branch_output, builder, 'index.html')

    if os.path.exists(search_page):
        copy_if_needed(source_file=search_page,
                       target_file=os.path.join(single_html_dir, 'search.html'))

    dest = os.path.join(conf.paths.projectroot, conf.paths.public_site_output)

    cmd_str = 'rsync -a {source}/ {destination}'.format(source=sconf.fq_build_output,
                                                        destination=dest)

    with open(os.devnull, 'w') as f:
        return_code = subprocess.call(args=cmd_str.split(),
                                      stdout=f,
                                      stderr=f)
        m = '"{0}" migrated build from {1} to {2}, with result {3}'
        logger.info(m.format(sconf.name, sconf.fq_build_output, dest, return_code))

    if 'excluded_files' in sconf:
        fns = [os.path.join(conf.paths.projectroot,
                            conf.paths.public_site_output,
                            fn)
               for fn in sconf['dirhtml']['excluded_files']]

        for fn in fns:
            if os.path.isdir(fn):
                shutil.rmtree(fn)
            elif os.path.isfile(fn):
                os.remove(fn)
            else:
                continue

            logger.info('removed file from dirhtml output directory: ' + fn)

    if conf.git.branches.current in conf.git.branches.published:
        sitemap_exists = sitemap(config_path=None, conf=conf)

        legacy_sitemap_fn = os.path.join(conf.paths.projectroot,
                                         conf.paths.branch_output,
                                         'sitemap.xml.gz')

        if os.path.exists(legacy_sitemap_fn) and sitemap_exists is True:
            copy_if_needed(source_file=legacy_sitemap_fn,
                           target_file=os.path.join(conf.paths.projectroot,
                                                    conf.paths.public_site_output,
                                                    'sitemap.xml.gz'))


def sitemap(config_path, conf):
    sys.path.append(os.path.join(conf.paths.projectroot, conf.paths.buildsystem, 'bin'))
    import sitemap_gen

    config_paths = []
    if config_path is not None:
        config_paths.append(config_path)
        config_paths.append(os.path.join(conf.paths.projectroot, conf.paths.builddata, config_path))
        config_path = None

    default_name = 'conf-sitemap.xml'
    config_paths.extend([default_name,
                         os.path.join(conf.paths.projectroot, default_name),
                         os.path.join(conf.paths.projectroot,
                                      conf.paths.builddata, default_name)])

    if 'edition' in conf.project and conf.project.edition != conf.project.name:
        edition_name = '-'.join(['conf', conf.project.edition, 'sitemap.xml'])
        config_paths.extend([edition_name,
                            os.path.join(conf.paths.projectroot, edition_name),
                            os.path.join(conf.paths.projectroot,
                                         conf.paths.builddata, edition_name)])

    for path in config_paths:
        if os.path.isfile(path):
            config_path = path
            break

    if config_path is None:
        m = 'sitemap: configuration file {0} does not exist. Returning early'
        logger.error(m.format(config_path))
        return False

    sitemap = sitemap_gen.CreateSitemapFromFile(configpath=config_path,
                                                suppress_notify=True)
    if sitemap is None:
        logger.error('sitemap: failed to generate the sitemap due to encountered errors.')
        return False

    sitemap.Generate()

    logger.info('sitemap: generated sitemap according to the config file {0}'.format(config_path))
    return True
